class Player:
    def __init__(self, nick):
        self.nick = nick
        self.values = {}

    def get(self, key):
        try:
            return self.values[key]
        except KeyError:
            return ""

    def set(self, key, value):
        self.values[key] = value

    def __str__(self):
        return "<Player {0}>".format(self.nick)



import fsbc.settings as settings
from arcade.resources import resources
from fsgs.option import Option


class ArcadeTheme:
    __instance = None

    @classmethod
    def get(cls):
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self):
        self.theme = settings.get(Option.ARCADE_THEME)
        if not self.theme:
            self.theme = "blue"

    def qimage(self, name):
        if self.theme != "blue":
            theme_path = "themes/{}/{}".format(self.theme, name)
            try:
                im = resources.resource_qt_image(theme_path)
            except LookupError:
                pass
            else:
                return im
        theme_path = "themes/blue/{}".format(name)
        im = resources.resource_qt_image(theme_path)
        return im



import struct


class RomAccess:
  def __init__(self, rom_data):
    self.rom_data = rom_data
    if type(rom_data) is bytes:
      self.writable = False
    elif type(rom_data) is bytearray:
      self.writable = True
    else:
      raise ValueError("rom_data must be bytes or bytearray")
    self.size = len(rom_data)
    self.kib = int(self.size / 1024)

  def get_size(self):
    return self.size

  def get_size_kib(self):
    return self.kib

  def is_writable(self):
    return self.writable

  def get_data(self):
    return self.rom_data

  def make_writable(self):
    if not self.writable:
      self.writable = True
      self.rom_data = bytearray(self.rom_data)
    return self.rom_data

  def read_long(self, off):
    return struct.unpack_from(">I", self.rom_data, off)[0]

  def write_long(self, off, val):
    if not self.writable:
      raise ValueError("can't write to ROM")
    return struct.pack_into(">I", self.rom_data, off, val)

  def read_word(self, off):
    return struct.unpack_from(">H", self.rom_data, off)[0]

  def write_word(self, off, val):
    if not self.writable:
      raise ValueError("can't write to ROM")
    return struct.pack_into(">H", self.rom_data, off, val)

  def read_byte(self, off):
    return struct.unpack_from("B", self.rom_data, off)[0]

  def read_sbyte(self, off):
    return struct.unpack_from("b", self.rom_data, off)[0]



import os
import sys

import fsboot
from fsbc.path import unicode_path
from fsbc.util import memoize

CSIDL_DESKTOP = 0
CSIDL_PERSONAL = 5
CSIDL_APPDATA = 26
CSIDL_MYPICTURES = 39
CSIDL_PROFILE = 40
CSIDL_COMMON_DOCUMENTS = 46


@memoize
def get_user_name():
    return fsboot.user_name()


@memoize
def xdg_user_dir(name):
    return fsboot.xdg_user_dir(name)


@memoize
def get_desktop_dir(allow_create=True):
    if sys.platform == "win32":
        path = fsboot.csidl_dir(fsboot.CSIDL_DESKTOP)
    else:
        path = xdg_user_dir("DESKTOP")
        if not path:
            path = os.path.join(get_home_dir(), "Desktop")
    path = unicode_path(path)
    if allow_create and not os.path.isdir(path):
        os.makedirs(path)
    return path


def get_documents_dir(create=False):
    return fsboot.documents_dir(create=create)


@memoize
def get_common_documents_dir():
    if sys.platform == "win32":
        path = fsboot.csidl_dir(CSIDL_COMMON_DOCUMENTS)
    else:
        raise NotImplementedError("Only for windows")
    path = unicode_path(path)
    return path


@memoize
def get_pictures_dir(allow_create=True):
    if sys.platform == "win32":
        path = fsboot.csidl_dir(CSIDL_MYPICTURES)
    else:
        path = xdg_user_dir("PICTURES")
        if not path:
            path = os.path.join(get_home_dir(), "Pictures")
    path = unicode_path(path)
    if allow_create and not os.path.isdir(path):
        os.makedirs(path)
    return path


def get_home_dir():
    return fsboot.home_dir()


def get_data_dir(create=True):
    return fsboot.common_data_dir(create=create)



# noinspection PyPep8Naming
def SDL_Maximize():
    pass


# noinspection PyPep8Naming
def SDL_Restore():
    pass


# noinspection PyPep8Naming
def SDL_Minimize():
    pass


# noinspection PyPep8Naming
def SDL_IsMaximized():
    return False


# noinspection PyPep8Naming
def SDL_IsMinimized():
    return False



import re
import subprocess
import traceback

from fsbc.system import macosx, windows

if windows:
    # noinspection PyUnresolvedReferences
    try:
        import win32api
    except ImportError:
        win32api = None
    # noinspection PyUnresolvedReferences
    try:
        import win32con
    except ImportError:
        win32con = None
    EDS_RAWMODE = 2
elif macosx:
    # noinspection PyUnresolvedReferences
    try:
        import Quartz  # type: ignore
    except ImportError:
        # Some people install unofficial newer macOS versions on older hardware, and that breaks
        # Quartz / Coregraphics (https://github.com/FrodeSolheim/fs-uae-launcher/issues/154).
        # Easy workaround applied, since Quartz isn't essential.
        traceback.print_exc()
        Quartz = None


class RefreshRateTool(object):
    def __init__(self, game_platform=None, game_refresh=None):
        self.game_platform = game_platform
        self.game_refresh = game_refresh

    def set_best_mode(self):
        print("RefreshRateTool.set_best_mode")
        current = self.get_current_mode()
        print("REFRESH RATE CONTROL: current mode is", current)
        modes = self.get_all_modes()
        mode_score_list = []
        for mode in modes:
            # print(mode)
            # for now, only consider same resolution
            if mode["width"] != current["width"]:
                continue
            if mode["height"] != current["height"]:
                continue
            # only consider modes with the same bpp
            if mode["bpp"] != current["bpp"]:
                continue
            score = self.calculate_mode_score(mode)
            if score is not None:
                mode_score_list.append((score, mode))
        if len(mode_score_list) == 0:
            print("REFRESH RATE TOOL: mode_score_list is empty")
            return
        print("REFRESH RATE TOOL:", mode_score_list)
        best_mode = mode_score_list[0][1]
        if best_mode == current:
            print("REFRESH RATE TOOL: using current mode")
        else:
            self.set_mode(best_mode)

    def calculate_mode_score(self, mode, debug=False):
        game_refresh = self.game_refresh
        if not self.game_refresh:
            if debug:
                print("REFRESH RATE TOOL: game refresh not specified")
            return None
        # in case of values such as -1 etc
        if game_refresh < 1:
            if debug:
                print("REFRESH RATE TOOL: game refresh not specified")
            return None
        display_refresh = mode["refresh"]
        if not display_refresh:
            if debug:
                print("REFRESH RATE TOOL: display refresh not found")
            return None
        # in case of values such as -1 etc
        if display_refresh < 1:
            if debug:
                print("REFRESH RATE TOOL: display refresh not found")
            return None
        diff = game_refresh - display_refresh
        # FIXME: configurable?
        # FIXME: configurable per platform?
        allowable_pos_diff = 1.01
        allowable_neg_diff = 1.01
        if debug:
            print(
                "REFRESH RATE TOOL: game refresh: {0} vs "
                "display refresh: {1}".format(game_refresh, display_refresh)
            )
            print(
                "REFRESH RATE TOOL: allow +{0}/-{1}".format(
                    allowable_pos_diff, allowable_neg_diff
                )
            )
            print("REFRESH RATE TOOL: diff: {0}".format(diff))
        if allowable_pos_diff > diff > -allowable_neg_diff:
            if debug:
                print("REFRESH RATE TOOL: allow vsync")
            return diff
        if debug:
            print("REFRESH RATE TOOL: deny vsync")
        return None

    def allow_vsync(self):
        """
        game_refresh = self.game_refresh
        if not self.game_refresh:
            print("REFRESH RATE TOOL: game refresh not specified")
            return False
        # in case of values such as -1 etc
        if game_refresh < 1:
            print("REFRESH RATE TOOL: game refresh not specified")
            return False
        display_refresh = self.get_display_refresh()
        if not display_refresh:
            print("REFRESH RATE TOOL: display refresh not found")
            return False
        # in case of values such as -1 etc
        if display_refresh < 1:
            print("REFRESH RATE TOOL: display refresh not found")
            return False
        diff = game_refresh - display_refresh
        # FIXME: configurable?
        # FIXME: configurable per platform?
        allowable_pos_diff = 1.01
        allowable_neg_diff = 1.01
        print("REFRESH RATE TOOL: game refresh: {0} vs "
                "display refresh: {1}".format(game_refresh,
                display_refresh))
        print("REFRESH RATE TOOL: allow +{0}/-{1}".format(allowable_pos_diff,
                allowable_neg_diff))
        print("REFRESH RATE TOOL: diff: {0}".format(diff))
        if diff < allowable_pos_diff and diff > -allowable_neg_diff:
            print("REFRESH RATE TOOL: allow vsync")
            return True
        print("REFRESH RATE TOOL: deny vsync")
        """
        current = self.get_current_mode()
        return self.calculate_mode_score(current, debug=True) is not None

    def get_display_refresh(self):
        # # FIXME:
        # if macosx:
        #     return 60.0
        # elif windows:
        #     import win32api
        #     device = win32api.EnumDisplayDevices()
        #     settings = win32api.EnumDisplaySettings(device.DeviceName, 0)
        #     return settings.DisplayFrequency()
        # # FIXME:
        # return 50.0
        # return None
        return self.get_current_mode()["refresh"]

    def get_current_mode(self):
        refresh = 0
        width = 640
        height = 480
        bpp = 0
        flags = 0

        if windows:
            if win32api:
                settings = win32api.EnumDisplaySettings(
                    None, win32con.ENUM_CURRENT_SETTINGS
                )
                refresh = float(settings.DisplayFrequency)
                width = int(settings.PelsWidth)
                height = int(settings.PelsHeight)
                bpp = int(settings.BitsPerPel)
                flags = int(settings.DisplayFlags)
        elif macosx:
            if Quartz is None:
                print("WARNING: Quartz is not available, using fake values")
                width = 640
                height = 480
                refresh = 0
            else:
                main_display = Quartz.CGMainDisplayID()
                current_display = Quartz.CGDisplayPrimaryDisplay(main_display)
                current_mode = Quartz.CGDisplayCopyDisplayMode(current_display)
                width = Quartz.CGDisplayModeGetWidth(current_mode)
                height = Quartz.CGDisplayModeGetHeight(current_mode)
                refresh = Quartz.CGDisplayModeGetRefreshRate(current_mode)

            if not refresh:
                print("WARNING: returned refresh rate was 0. assuming 60 Hz")
                refresh = 60
            # A bit weird that it crashes, since this is supposed to be a
            # copy of the mode...?
            print(
                "FIXME: Not calling Quartz.CGDisplayModeRelease("
                "current_mode), seems to crash pygame on exit..."
            )
            # Quartz.CGDisplayModeRelease(current_mode)
            flags = 0
            bpp = None
        else:
            return self._get_current_mode_x()

        return {
            "width": width,
            "height": height,
            "refresh": refresh,
            "bpp": bpp,
            "flags": flags,
        }

    def get_all_modes(self):
        modes = []
        if windows:
            k = 0
            while True:
                try:
                    settings = win32api.EnumDisplaySettingsEx(
                        None, k, EDS_RAWMODE
                    )
                except win32api.error:
                    break
                refresh = float(settings.DisplayFrequency)
                width = int(settings.PelsWidth)
                height = int(settings.PelsHeight)
                bpp = int(settings.BitsPerPel)
                flags = int(settings.DisplayFlags)
                # print(width, height, refresh, bpp, flags)
                modes.append(
                    {
                        "width": width,
                        "height": height,
                        "refresh": refresh,
                        "bpp": bpp,
                        "flags": flags,
                    }
                )
                k += 1
        elif macosx:
            # FIXME:
            pass
        else:
            modes = self._get_all_modes_x()
        # modes.extend(self.get_override_modes())
        return modes

    # def get_override_modes(self):
    #     modes = []
    #     k = 0
    #     while True:
    #         mode_string = pyapp.user.ini.get('ForceDisplayModes/%d' % k)
    #         if not mode_string:
    #             break
    #         mode_string = mode_string.lower()
    #         mode = {}
    #         w, rest = mode_string.split('x')
    #         h, rest = rest.split('@')
    #         r, rest = rest.split('hz')
    #         mode['width'] = int(w)
    #         mode['height'] = int(h)
    #         mode['bpp'] = 32
    #         mode['refresh'] = int(r)
    #         mode['flags'] = 0
    #         modes.append(mode)
    #         k += 1
    #     print("get_override_modes returning", modes)
    #     return modes

    def set_mode(self, mode):
        # if windows:
        #     self._set_mode_windows(mode)
        # elif macosx:
        #     # FIXME:
        #     print("WARNING: mode settings is not supported on this "
        #             "platform yet")
        # else:
        #     self._set_mode_x(mode)

        print("FIXME: Currently disabled set_mode")
        return

    def _set_mode_windows(self, mode):
        print("REFRESH RATE TOOL: setting mode", mode)
        if not win32api:
            print("win32api not available, returning")
            return False
        k = 0
        while True:
            try:
                settings = win32api.EnumDisplaySettingsEx(None, k, EDS_RAWMODE)
            except win32api.error:
                break
            refresh = float(settings.DisplayFrequency)
            width = int(settings.PelsWidth)
            height = int(settings.PelsHeight)
            bpp = int(settings.BitsPerPel)
            flags = int(settings.DisplayFlags)
            # print(width, height, refresh, bpp, flags)
            # modes.append({'width': width, 'height': height,
            #        'refresh': refresh, 'bpp': bpp, 'flags': flags})
            if (
                width == mode["width"]
                and height == mode["height"]
                and refresh == mode["refresh"]
                and bpp == mode["bpp"]
                and flags == mode["flags"]
            ):
                # print("trying to override with refresh", int(round(self.game_refresh)))
                # #refresh == mode['refresh'] and \
                # settings.DisplayFrequency = int(round(self.game_refresh))
                # result = win32api.ChangeDisplaySettings(settings,
                #         win32con.CDS_UPDATEREGISTRY) #win32con.CDS_FULLSCREEN)
                #         #0) #win32con.CDS_FULLSCREEN)
                # if result == win32con.DISP_CHANGE_SUCCESSFUL:
                #     print("display change was successful")
                #     return True
                # print("failed, falling back to ", mode['refresh'])
                # settings.DisplayFrequency = mode['refresh']
                print("found windows mode, changing display settings")
                result = win32api.ChangeDisplaySettings(
                    settings, win32con.CDS_UPDATEREGISTRY
                )
                # win32con.CDS_FULLSCREEN)

                if result == win32con.DISP_CHANGE_SUCCESSFUL:
                    print("display change was successful")
                    return True
                else:
                    print("display change failed, result =", result)
                    return False
            k += 1

        # for omode in self.get_override_modes():
        #     print("trying override mode", omode)
        #     if omode['width'] == mode['width'] and \
        #             omode['height'] == mode['height'] and \
        #             omode['refresh'] == mode['refresh'] and \
        #             omode['bpp'] == mode['bpp'] and \
        #             omode['flags'] == mode['flags']:
        #         settings.PelsWidth = omode['width']
        #         settings.PelsHeight = omode['height']
        #         settings.BitsPerPel = omode['bpp']
        #         settings.DisplayFlags = omode['flags']
        #         settings.DisplayFrequency = omode['refresh']
        #         result = win32api.ChangeDisplaySettings(settings,
        #                 win32con.CDS_UPDATEREGISTRY) #win32con.CDS_FULLSCREEN)
        #                 #0) #win32con.CDS_FULLSCREEN)
        #         if result == win32con.DISP_CHANGE_SUCCESSFUL:
        #             print("display change was successful")
        #             return True
        #         else:
        #             print("display change failed, result =", result)
        #             return False
        return False

    def _set_mode_x(self, mode):
        print("REFRESH RATE TOOL: setting mode", mode)
        args = [
            "/usr/bin/env",
            "xrandr",
            "-s",
            "{0}x{1}".format(mode["width"], mode["height"]),
            "-r",
            str(mode["refresh"]),
        ]
        p = subprocess.Popen(args)
        p.wait()

    def _get_all_modes_x(self):
        modes = []
        self._get_current_mode_x(modes=modes)
        return modes

    def _get_current_mode_x(self, modes=[]):
        modes[:] = []
        mode = {"width": 0, "height": 0, "refresh": 0.0, "bpp": 0, "flags": 0}
        args = ["/usr/bin/env", "xrandr", "-q"]
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        for line in p.stdout:
            line = line.decode("ISO-8859-1")
            if not line.startswith(" "):
                continue
            line = line.strip()
            line = line.replace("+", "")
            # collapse multiple spaces
            line = re.sub(" +", " ", line).strip()
            parts = line.split(" ")
            resolution = parts[0]
            rates = parts[1:]
            width, height = resolution.split("x")
            width = int(width)
            height = int(height)
            print(rates)
            for rate in rates:
                if rate[-1] == "*":
                    refresh = float(rate[:-1])
                    mode["width"] = width
                    mode["height"] = height
                    mode["refresh"] = refresh
                    modes.append(
                        {
                            "width": width,
                            "height": height,
                            "refresh": float(rate[:-1]),
                            "bpp": 0,
                            "flags": 0,
                        }
                    )
                else:
                    modes.append(
                        {
                            "width": width,
                            "height": height,
                            "refresh": float(rate),
                            "bpp": 0,
                            "flags": 0,
                        }
                    )
        return mode

    def screens_xrandr(self):
        try:
            return self._screens_xrandr()
        except Exception:
            traceback.print_exc()
            return {}

    @staticmethod
    def _screens_xrandr():
        screens = {}
        screen = {"modes": []}
        args = ["/usr/bin/env", "xrandr", "-q"]
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        data = p.stdout.read().decode("ISO-8859-1")
        # data = XRANDR_TEST_DATA
        for line in data.split("\n"):
            last_refresh_rate = 0.0
            if line.startswith(" "):
                line = line.replace("*", " * ")
                line = line.replace("+", " + ")
                line = re.sub(" +", " ", line).strip()
                parts = line.split(" ")
                # May not always have a resolution here, could be
                # a named mode instead
                # resolution_str = parts[0]
                # w, h = resolution_str.split("x")
                # w = int(w)
                # h = int(h)
                for part in parts[1:]:
                    if part == "*":
                        screen["refresh_rate"] = last_refresh_rate
                    elif part == "+":
                        pass
                    else:
                        last_refresh_rate = float(part)
                        # screen["modes"].append({
                        #     "width": w,
                        #     "height": h,
                        #     "refresh_rate": last_refresh_rate
                        # })
                        pass
            else:
                line = re.sub(" +", " ", line).strip()
                parts = line.split(" ")
                for part in parts:
                    if "x" not in part or "+" not in part:
                        continue
                    geom = part.replace("x", " ").replace("+", " ")
                    print(geom)
                    w, h, x, y = geom.split(" ")
                    w = int(w)
                    h = int(h)
                    x = int(x)
                    y = int(y)
                    screen = {
                        "x": x,
                        "y": x,
                        "width": x,
                        "height": x,
                        "modes": [],
                        "refresh_rate": 0.0,
                    }
                    screens[(x, y, w, h)] = screen
        print(screens)
        return screens


# noinspection SpellCheckingInspection
XRANDR_TEST_DATA = """\
Screen 0: minimum 8 x 8, current 1920 x 1080, maximum 32767 x 32767
DP1 disconnected (normal left inverted right x axis y axis)
DP2 disconnected (normal left inverted right x axis y axis)
DP3 disconnected (normal left inverted right x axis y axis)
HDMI1 connected primary 1920x1080+0+0 (normal left inverted right x \
axis y axis) 531mm x 298mm
   1920x1080     60.00*+
   Amiga         50.02
   1680x1050     59.88
   1600x900      60.00
   1280x1024     75.02    60.02
   1280x800      59.91
   1152x864      75.00
   1280x720      60.00
   1024x768      75.08    60.00
   832x624       74.55
   800x600       75.00    60.32
   640x480       75.00    60.00
   720x400       70.08
HDMI2 disconnected (normal left inverted right x axis y axis)
HDMI3 disconnected (normal left inverted right x axis y axis)
VGA1 connected (normal left inverted right x axis y axis)
   1680x1050     59.95 +
   AmigaVGA      49.97
   1600x1000     60.01
   1280x1024     75.02
   1440x900      59.89
   1280x960      60.00
   1152x864      75.00
   1152x720      59.97
   1024x768      75.08    60.00
   832x624       74.55
   800x600       75.00    60.32
   640x480       75.00    60.00
   720x400       70.08
VIRTUAL1 disconnected (normal left inverted right x axis y axis)
"""



import hashlib
import os
import shutil
from uuid import NAMESPACE_URL, uuid4, uuid5

import requests

from fsbc.application import app
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.network import fs_uae_url_from_sha1_and_name
from fsgs.option import Option
from fsgs.plugins.pluginmanager import PluginManager


class OfflineModeException(Exception):
    pass


def offline_mode():
    return app.settings[Option.OFFLINE_MODE] == "1"


def raise_exception_in_offline_mode():
    if offline_mode():
        raise OfflineModeException("Offline mode is enabled")


class Downloader(object):
    @classmethod
    def check_terms_accepted(cls, download_file, download_terms):
        print(
            "[DOWNLOADER] check_terms_accepted", download_file, download_terms
        )
        uuid = str(uuid5(NAMESPACE_URL, download_file + download_terms))
        path = cls.get_cache_path(uuid)
        print("[DOWNLOADER] check_terms_accepted", path)
        return os.path.exists(path)

    @classmethod
    def set_terms_accepted(cls, download_file, download_terms):
        print(
            "[DOWNLOADER] set_terms_accepted",
            repr(download_file),
            repr(download_terms),
        )
        uuid = str(uuid5(NAMESPACE_URL, download_file + download_terms))
        path = cls.get_cache_path(uuid)
        print("[DOWNLOADER] set_terms_accepted", path)
        with open(path, "wb") as _:
            pass

    @classmethod
    def cache_data(cls, data):
        sha1 = hashlib.sha1(data).hexdigest()
        cache_path = cls.get_cache_path(sha1)
        if os.path.exists(cache_path):
            return
        cache_path_temp = cache_path + ".partial." + str(uuid4())
        with open(cache_path_temp, "wb") as f:
            f.write(data)
        try:
            os.rename(cache_path_temp, cache_path)
        except Exception as e:
            if os.path.exists(cache_path):
                # file has appeared in the meantime (by another thread?)
                return
            raise e

    @classmethod
    def cache_file_from_url(cls, url, download=True, auth=None):
        print("[DOWNLOADER] cache_file_from_url", url)
        cache_path = cls.get_url_cache_path(url)
        if os.path.exists(cache_path):
            print("[DOWNLOADER] (in cache)")
            # so we later can delete least accessed files in cache...
            os.utime(cache_path, None)
            return cache_path
        if not download:
            return None
        raise_exception_in_offline_mode()
        r = requests.get(url, auth=auth, stream=True)
        try:
            r.raise_for_status()
            cache_path_temp = cache_path + ".partial." + str(uuid4())
            with open(cache_path_temp, "wb") as ofs:
                for chunk in r.iter_content(chunk_size=65536):
                    ofs.write(chunk)
        finally:
            r.close()
        os.rename(cache_path_temp, cache_path)
        return cache_path

    @classmethod
    def install_file_from_url(cls, url, path):
        print("[DOWNLOADER] install_file_from_url", url)
        print(repr(path))
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        cache_path = cls.cache_file_from_url(url)
        temp_path = path + ".partial"
        shutil.copy(cache_path, temp_path)
        os.rename(temp_path, path)

    @classmethod
    def install_file_by_sha1(cls, sha1, name, path):
        print("[DOWNLOADER] install_file_by_sha1", sha1)
        # FIXME: Also find files from file database / plugins

        print(repr(path))
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        src = PluginManager.instance().find_file_by_sha1(sha1)
        if src:
            dst = path
            dst_partial = dst + ".partial"
            sha1_obj = hashlib.sha1()
            with open(src, "rb") as fin:
                with open(dst_partial, "wb") as fout:
                    while True:
                        data = fin.read(65536)
                        if not data:
                            break
                        fout.write(data)
                        sha1_obj.update(data)
            if sha1_obj.hexdigest() != sha1:
                raise Exception("File from plugin does not match SHA-1")
            os.rename(dst_partial, dst)
            return

        cache_path = cls.get_cache_path(sha1)
        if os.path.exists(cache_path):
            print("[CACHE]", cache_path)
            # FIXME: Atomic copy utility function?
            shutil.copy(cache_path, path)
            # so we later can delete least accessed files in cache...
            os.utime(cache_path, None)
            return
        url = cls.sha1_to_url(sha1, name)
        print("[DOWNLOADER]", url)
        raise_exception_in_offline_mode()
        r = requests.get(url, stream=True)
        try:
            r.raise_for_status()
            temp_path = path + ".partial." + str(uuid4())
            h = hashlib.sha1()
            with open(temp_path, "wb") as output:
                for chunk in r.iter_content(chunk_size=65536):
                    h.update(chunk)
                    output.write(chunk)
        finally:
            r.close()
        if h.hexdigest() != sha1:
            print("error: downloaded sha1 is", h.hexdigest(), "- wanted", sha1)
            raise Exception("sha1 of downloaded file does not match")

        # Atomic "copy" to cache location
        temp_cache_path = cache_path + ".partial." + str(uuid4())
        shutil.copy(temp_path, temp_cache_path)
        os.rename(temp_cache_path, cache_path)

        # Move downloaded file into file position (atomic)
        os.rename(temp_path, path)

    @classmethod
    def sha1_to_url(cls, sha1, name):
        return fs_uae_url_from_sha1_and_name(sha1, name)

    @classmethod
    def get_cache_path(cls, sha1_or_uuid):
        path = os.path.join(
            FSGSDirectories.get_cache_dir(), "Downloads", sha1_or_uuid[:3]
        )
        if not os.path.exists(path):
            os.makedirs(path)
        return os.path.join(path, sha1_or_uuid)

    @classmethod
    def get_url_cache_path(cls, url):
        return cls.get_cache_path(str(uuid5(NAMESPACE_URL, url)))



import os
import zipfile

from .application import Application


class Resources(object):
    zip_files = {}

    @classmethod
    def resource_zip(cls, package: str):
        application = Application.get_instance()
        try:
            file = application.data_file(f"{package}.zip")
        except LookupError:
            file = None
        if package not in cls.zip_files:
            if file is None:
                cls.zip_files[package] = None
            else:
                zip_file = zipfile.ZipFile(file)
                cls.zip_files[package] = zip_file
        return cls.zip_files[package]

    def __init__(self, package, subdir=""):
        self.package = package
        self.subdir = subdir
        self.resource_sha1s = {}

    def resource_name(self, resource):
        if self.subdir:
            # return os.path.join(self.subdir, resource)
            return self.subdir + "/" + resource
        return resource

    def stream(self, resource):
        print("Resources.stream", self.package, resource)
        zip = self.resource_zip(self.package)
        if zip is not None:
            resource_name = self.resource_name(resource)
            print("Resources zip.open", zip, resource_name)
            try:
                return zip.open(resource_name)
            except KeyError:
                pass

        try:
            return open(self.path(resource), "rb")
        except FileNotFoundError:
            raise LookupError(resource)

    def path(self, resource):
        application = Application.get_instance()
        resource_name = self.resource_name(resource)
        print("looking up resource", resource_name, "package =", self.package)

        relative_path = os.path.join(self.package, resource_name)
        try:
            return application.data_file(relative_path)
        except LookupError:
            pass

        raise LookupError("Cannot find resource {0}".format(repr(resource)))



"""Helper functions and constants useable with DosTypes"""

# raw dos types
DOS0 = 0x444f5300
DOS1 = 0x444f5301
DOS2 = 0x444f5302
DOS3 = 0x444f5303
DOS4 = 0x444f5304
DOS5 = 0x444f5305
DOS6 = 0x444f5306
DOS7 = 0x444f5307

# more convenient dos type
DOS_OFS = DOS0
DOS_FFS = DOS1
DOS_OFS_INTL = DOS2
DOS_FFS_INTL = DOS3
DOS_OFS_INTL_DIRCACHE = DOS4
DOS_FFS_INTL_DIRCACHE = DOS5
DOS_OFS_INTL_LONGNAME = DOS6
DOS_FFS_INTL_LONGNAME = DOS7

# string names for dos types
dos_type_names = [
    'DOS0:ofs',
    'DOS1:ffs',
    'DOS2:ofs+intl',
    'DOS3:ffs+intl',
    'DOS4:ofs+intl+dircache',
    'DOS5:ffs+intl+dircache',
    'DOS6:ofs+intl+longname',
    'DOS7:ffs+intl+longname'
]

# masks for modes
DOS_MASK_FFS = 1
DOS_MASK_INTL = 2
DOS_MASK_DIRCACHE = 4

def parse_dos_type_str(string):
  """parse a dos type string
     return None if its invalid or dostype value
  """
  comp = string.split("+")
  if "ffs" in comp:
    if "dc" in comp or "dircache" in comp:
      return DOS_FFS_INTL_DIRCACHE
    elif "intl" in comp:
      return DOS_FFS_INTL
    else:
      return DOS_FFS
  elif "ofs" in comp:
    if "dc" in comp or "dircache" in comp:
      return DOS_OFS_INTL_DIRCACHE
    elif "intl" in comp:
      return DOS_OFS_INTL
    else:
      return DOS_OFS
  else:
    n = len(string)
    # use 'DOS0' .. 'DOS5'
    if n == 4 and string[0:3] == 'DOS':
      off = ord(string[3]) - ord('0')
      if off >= 0 and off <= 7:
        return DOS0 + off
      else:
        return None
    # other tag?
    elif string[0].isalpha() and n==4:
      return tag_str_to_num(string)
    # use '0x01234567' hex value
    elif string[0:2] == '0x':
      try:
        return int(string[2:],16)
      except ValueError:
        return None
    # try number
    else:
      try:
        return int(string)
      except ValueError:
        return None

def tag_str_to_num(s):
  """Convert the DosType in a 4 letter tag string to 32 bit value"""
  if len(s) != 4:
    return 0
  a = ord(s[0]) << 24
  b = ord(s[1]) << 16
  c = ord(s[2]) << 8
  l = s[3]
  d = ord(l)
  if l.isdigit():
    d = d - ord('0')
  return a+b+c+d

def num_to_tag_str(l):
  """Convert the DosType in a 32 bit value to its 4 letter tag string"""
  a = chr((l >> 24) & 0xff)
  b = chr((l >> 16) & 0xff)
  c = chr((l >> 8) & 0xff)
  last = (l & 0xff)
  if last < 32:
    last = chr(last + 48)
  else:
    last = chr(last)
  return a+b+c+last

def get_dos_type_str(dos_type):
  """return description of dos type as a string"""
  return dos_type_names[dos_type & 0x7]

def is_valid(dos_type):
  """check if its a valid dos type"""
  return (dos_type >= DOS0) and (dos_type <= DOS7)

def is_ffs(dos_type):
  """check if its a fast file system dostype"""
  return (dos_type & DOS_MASK_FFS) == DOS_MASK_FFS

def is_intl(dos_type):
  """check if international mode is enabled in dostype"""
  return is_dircache(dos_type) or is_longname(dos_type) or (dos_type & DOS_MASK_INTL) == DOS_MASK_INTL

def is_dircache(dos_type):
  """check if dir cache mode is enabled in dostype"""
  return (dos_type == DOS4) or (dos_type == DOS5)

def is_longname(dos_type):
  """check if long filename mode is enabled in dostype"""
  return (dos_type == DOS6) or (dos_type == DOS7)

def rootblock_tracks_used_blocks(dos_type):
  """checks if the number of used blocks is stored within the rootblock"""
  return (dos_type == DOS6) or (dos_type == DOS7)



import fsui

from ...game_paths import GamePaths
from ...launcher_config import LauncherConfig
from ...launcher_settings import LauncherSettings
from ..Constants import Constants
from ..ImageLoader import ImageLoader
from ..skin import LauncherTheme, Skin
from .BottomPanel import BottomPanel
from .EditButton import EditButton
from .WebButton import WebButton

BORDER = 20


class GameInfoPanel(BottomPanel):
    def __init__(self, parent, cover_on_right=True):
        BottomPanel.__init__(self, parent)
        Skin.set_background_color(self)
        # self.set_background_color((0xdd, 0xdd, 0xdd))

        self.cover_on_right = cover_on_right
        self.requests = None
        self.image = None

        self.layout = fsui.HorizontalLayout()

        # def get_min_height():
        #     return Constants.COVER_SIZE[1] + 2 * BORDER
        # self.layout.get_min_height = get_min_height

        self.layout.padding_left = BORDER // 2
        self.layout.padding_right = BORDER // 2
        self.layout.padding_top = BORDER + 2
        self.layout.padding_bottom = Skin.get_bottom_margin()

        self.default_image = fsui.Image("launcher:res/cover.png")
        # self.default_image.resize(Constants.COVER_SIZE)
        self.cover_overlay = fsui.Image("launcher:res/cover_overlay.png")
        self.cover_overlay_square = fsui.Image(
            "launcher:res/cover_overlay_square.png"
        )

        if not self.cover_on_right:
            self.layout.add_spacer(Constants.COVER_SIZE[0])

        # self.panel = fsui.Panel(self)
        # self.panel.set_background_color((0xdd, 0xdd, 0xdd))
        # self.layout.add(self.panel, expand=True, fill=True)

        vert_layout = fsui.VerticalLayout()
        self.layout.add(vert_layout, expand=True, fill=True)
        # self.panel.layout.padding_top = 10
        if self.cover_on_right:
            vert_layout.padding_right = BORDER
        else:
            vert_layout.padding_left = BORDER

        vert_layout.add_spacer(58 + 3)
        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        hori_layout.add_spacer(0, expand=True)

        self.web_button = WebButton(self)
        hori_layout.add(self.web_button, margin_right=3)

        self.edit_button = EditButton(self)
        hori_layout.add(self.edit_button, margin_right=3)

        # def label_min_width():
        #     return 330
        # self.title_label = fsui.HeadingLabel(self)
        # self.title_label.get_min_width = label_min_width
        # vert_layout.add(self.title_label)

        # self.sub_title_label = fsui.Label(self)
        # self.sub_title_label.get_min_width = label_min_width
        # vert_layout.add(self.sub_title_label)

        self.title = ""
        self.sub_title = ""
        self.year = ""
        self.publisher = ""
        self.developer = ""
        self.companies = ""

        self.image_path = ""

        vert_layout.add_spacer(0, expand=True)

        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        # for rating in [1, 4, 5]:
        #     button = RatingButton(self, rating)
        #     hori_layout.add(button, margin_right=5, fill=True)

        hori_layout.add_spacer(0, expand=True)
        # self.launch_group = LaunchGroup(self)
        # hori_layout.add(self.launch_group, fill=True)

        if self.cover_on_right:
            self.layout.add_spacer(Constants.COVER_SIZE[0])

        self.image_loader = ImageLoader()
        self.load_info()
        LauncherSettings.add_listener(self)
        LauncherConfig.add_listener(self)

        for key in [
            "database_url",
            "hol_url",
            "lemon_url",
            "mobygames_url",
            "wikipedia_url",
            "year",
            "publisher",
            "developer",
        ]:
            self.on_config(key, LauncherConfig.get(key))

    def on_destroy(self):
        LauncherConfig.remove_listener(self)
        LauncherSettings.remove_listener(self)
        self.image_loader.stop()

    def load_info(self):
        handler = GamePaths.current()
        name = handler.get_name()
        variant = handler.get_variant()
        if variant.startswith("Amiga, "):
            variant = variant[7:]
        self.title = name
        self.sub_title = variant
        self.sub_title = self.sub_title.replace(", ", " \u00b7 ")

        path = handler.get_cover_path()
        if path != self.image_path:
            self.image_path = path

            def on_load(request):
                if request.path != self.image_path:
                    return
                if request.image:
                    self.image = request.image
                else:
                    self.image = self.default_image
                self.image_path = request.path
                self.refresh()

            self.requests = self.image_loader.load_image(
                path, size=Constants.COVER_SIZE, on_load=on_load, is_cover=True
            )
            self.image = self.default_image

        self.refresh()
        self.layout.update()

    def on_setting(self, key, _):
        if key == "parent_uuid":
            self.load_info()
        if key == "config_name":
            self.load_info()

    def on_config(self, key, value):
        if key == "publisher":
            self.publisher = value
            self.update_companies()
        elif key == "developer":
            self.developer = value
            self.update_companies()
        elif key == "year":
            self.year = value

    def update_companies(self):
        companies = [x.strip() for x in self.publisher.split("/") if x.strip()]
        for developer in self.developer.split("/"):
            developer = developer.strip()
            if developer and developer not in companies:
                companies.append(developer)
        self.companies = " \u00b7 ".join(companies)

    def on_paint(self):
        dc = self.create_dc()
        dc.set_text_color(LauncherTheme.get().text_color)
        self.draw_background(dc)

        if self.cover_on_right:
            x = self.get_size()[0] - 10 - Constants.COVER_SIZE[0]
            text_x = 10
        else:
            x = 10
            text_x = 10 + Constants.COVER_SIZE[0] + 20

        y = 2 + 20
        size = self.size()

        image = self.image
        # dc.draw_image(image, x, y)
        if image.size[0] == image.size[1]:
            # cover_overlay = self.cover_overlay_square
            y_offset = 14
            # title_x = 10
        else:
            y_offset = 0
        cover_overlay = self.cover_overlay

        if image.size[0] == image.size[1]:
            dc.draw_rectangle(
                x + 1,
                y + 1,
                Constants.COVER_SIZE[0],
                Constants.COVER_SIZE[1],
                fsui.Color(0x0, 0x0, 0x0),
            )
        dc.draw_image(image, x + 1, y + 1 + y_offset)
        dc.draw_image(cover_overlay, x - 10, y - 10)

        font = dc.get_font()
        font.set_bold(True)
        dc.set_font(font)

        dc.draw_text(self.title, text_x, y)
        # tw, th = \
        dc.measure_text(self.title)
        # y += int(th * 1.2)
        y += 24

        color = dc.get_text_color()
        color.mix(self.get_background_color(), 0.33)
        dc.set_text_color(color)

        if self.year:
            twy, thy = dc.measure_text(self.year)
            dc.draw_text(self.year, text_x, y)
            twy += 10
        else:
            twy = 0
        font.set_bold(False)
        dc.set_font(font)
        if self.companies:
            dc.draw_text(self.companies, text_x + twy, y)

        background = fsui.Color(0x00, 0x00, 0x00, 0x20)
        y = 80
        h = 30

        tw, th = dc.measure_text(self.sub_title)
        if self.cover_on_right:
            dc.draw_rectangle(
                text_x,
                y,
                size[0] - Constants.COVER_SIZE[0] - 20 - 20,
                h,
                background,
            )
            dc.draw_text(self.sub_title, text_x + 6, y + (h - th) // 2)
        else:
            dc.draw_rectangle(
                text_x - 18, y, size[0] - text_x + 18, h, background
            )
            dc.draw_text(self.sub_title, text_x, y + (h - th) // 2)



import hashlib
import os

from fsgs import Option
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader


class GameGearPlatform(Platform):
    PLATFORM_NAME = "Game Gear"

    def driver(self, fsgc):
        return MednafenGameGearDriver(fsgc)

    def loader(self, fsgc):
        return GameGearLoader(fsgc)


class GameGearLoader(SimpleLoader):
    pass


class MednafenGameGearDriver(MednafenDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Built-in Controller",
        "mapping_name": "gamegear",
    }

    PORTS = [{"description": "Controller", "types": [CONTROLLER]}]

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = GameGearHelper(self.options)

    def prepare(self):
        super().prepare()
        self.set_mednafen_aspect(66, 50)
        # We do aspect calculation separately. Must not be done twice.
        # self.emulator.args.extend(["-snes.correct_aspect", "0"])
        rom_path = self.helper.prepare_rom(self)
        self.emulator.args.append(rom_path)

    def get_game_refresh_rate(self):
        return 59.73

    def mednafen_input_mapping(self, port):
        return {
            "A": "gb.input.builtin.gamepad.a",
            "B": "gb.input.builtin.gamepad.b",
            "UP": "gb.input.builtin.gamepad.up",
            "DOWN": "gb.input.builtin.gamepad.down",
            "LEFT": "gb.input.builtin.gamepad.left",
            "RIGHT": "gb.input.builtin.gamepad.right",
            "SELECT": "gb.input.builtin.gamepad.select",
            "START": "gb.input.builtin.gamepad.start",
        }

    def mednafen_rom_extensions(self):
        return [".gg"]

    def mednafen_scanlines_setting(self):
        return 0

    def mednafen_special_filter(self):
        return "nn2x"

    def mednafen_system_prefix(self):
        return "gg"

    def game_video_par(self):
        return (6.6 / 5.0) / (160 / 144)

    def game_video_size(self):
        return 160, 144

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class GameGearHelper:
    def __init__(self, options):
        self.options = options

    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    def prepare_rom_with_stream(self, driver, input_stream, ext):
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



import fsui
from fsgs.option import Option
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class LoggingSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("settings", "pkg:workspace")
        title = gettext("Logging")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "The following options may affect performance, "
                "so only enable them when needed for testing or "
                "debugging purposes."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=20)

        self.add_option(Option.LOG_AUTOSCALE)
        self.add_option(Option.LOG_INPUT)



from __future__ import absolute_import
from __future__ import print_function

class LogEntry:
  """A class for a log entry"""
  names = ('debug','info ','WARN ','ERROR')
  
  def __init__(self, level, msg, blk_num=-1):
    self.blk_num = blk_num
    self.level = level
    self.msg = msg
  def __str__(self):
    if self.blk_num == -1:
      return "%s%s:%s" % (" "*8, self.names[self.level], self.msg)
    else:
      return "@%06d:%s:%s" % (self.blk_num, self.names[self.level], self.msg)

class Log:
  """Store a log of entries"""

  DEBUG = 0
  INFO = 1
  WARN = 2
  ERROR = 3

  def __init__(self, min_level):
    self.entries = []
    self.min_level = min_level
  
  def msg(self, level, msg, blk_num = -1):
    if level < self.min_level:
      return
    e = LogEntry(level, msg, blk_num)
    self.entries.append(e)
  
  def dump(self):
    for e in self.entries:
      print(e)
  
  def get_num_level(self, level):
    num = 0
    for e in self.entries:
      if e.level == level:
        num += 1
    return num



from fsgs.platform import PlatformHandler
from fsgs.platforms.arcade.mamearcadedriver import MameArcadeDriver
from fsgs.platforms.loader import SimpleLoader


class ArcadePlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Arcade"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return ArcadeLoader(fsgs)

    def get_runner(self, fsgs):
        return MameArcadeDriver(fsgs)


class ArcadeLoader(SimpleLoader):
    def load_files(self, values):
        # file_list = json.loads(values["file_list"])
        # assert len(file_list) == 1
        # self.config["cartridge"] = "sha1://{0}/{1}".format(
        #     file_list[0]["sha1"], file_list[0]["name"])
        self.config["file_list"] = values["file_list"]

    def load_extra(self, values):
        if "refresh_rate" in values:
            self.config["refresh_rate"] = values["refresh_rate"]
        if "orientation" in values:
            self.config["orientation"] = values["orientation"]
        self.config["mame_rom_set"] = values["mame_rom_set"]



from fsgs.drivers.messdriver import MessDriver


class MessA2600Driver(MessDriver):
    CONTROLLER = {
        "type": "controller",
        "description": "Controller",
        "mapping_name": "atari2600",
    }

    PORTS = [
        {"description": "1st Controller", "types": [CONTROLLER]},
        {"description": "2nd Controller", "types": [CONTROLLER]},
    ]

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, port):
        return {
            "SELECT": 'tag=":SWB" type="OTHER" mask="2" defvalue="2"',
            "RESET": 'tag=":SWB" type="OTHER" mask="1" defvalue="1"',
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "1": "P#_BUTTON1",
            "2": "P#_BUTTON2",
            "3": "P#_BUTTON3",
        }

    def mess_romset(self):
        if self.is_pal():
            return "a2600p", {}
        else:
            return "a2600", {}



import fsui
from launcher.launcher_settings import LauncherSettings
from launcher.settings.option_ui import OptionUI
from launcher.settings.settings_header import SettingsHeader


class SettingsPage(fsui.Panel):
    def __init__(self, parent):
        super().__init__(parent)
        self.layout = fsui.VerticalLayout()
        self.layout.padding_top = 20
        self.layout.padding_bottom = 20
        self.layout.padding_left = 20
        self.layout.padding_right = 20
        self.icon_header = None
        self.options_on_page = set()

    def add_header(self, icon, title, subtitle=""):
        self.icon_header = SettingsHeader(self, icon, title, subtitle)
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

    def add_option(self, name, description=""):
        group = OptionUI.create_group(self, name, description=description)
        self.layout.add(group, fill=True, margin_top=10, margin_bottom=10)
        self.options_on_page.add(name)
        return group

    def add_section(self, title):
        label = fsui.HeadingLabel(self, title)
        self.layout.add(label, margin_top=20, margin_bottom=20)

    def reset_to_defaults(self):
        for option in self.options_on_page:
            LauncherSettings.set(option, "")



from fsgs import Option
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.knownfiles import KnownFile
from fsgs.platform import Platform, PlatformHandler
from fsgs.platforms.loader import SimpleLoader

PSX_PLATFORM_ID = "PSX"
PSX_PLATFORM_NAME = "PlayStation"
# noinspection SpellCheckingInspection
PSX_SCPH5500_BIN_SHA1 = "b05def971d8ec59f346f2d9ac21fb742e3eb6917"
# noinspection SpellCheckingInspection
PSX_SCPH5501_BIN_SHA1 = "0555c6fae8906f3f09baf5988f00e55f88e9f30b"
# noinspection SpellCheckingInspection
PSX_SCPH5502_BIN_SHA1 = "f6bc2d1f5eb6593de7d089c425ac681d6fffd3f0"
PSX_GAMEPAD = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "playstation",
}
PSX_SCPH5500_BIN = KnownFile(
    "b05def971d8ec59f346f2d9ac21fb742e3eb6917", PSX_PLATFORM_ID, "scph5500.bin"
)
PSX_SCPH5501_BIN = KnownFile(
    "0555c6fae8906f3f09baf5988f00e55f88e9f30b", PSX_PLATFORM_ID, "scph5501.bin"
)
# noinspection SpellCheckingInspection
PSX_SCPH5502_BIN = KnownFile(
    "f6bc2d1f5eb6593de7d089c425ac681d6fffd3f0", PSX_PLATFORM_ID, "scph5502.bin"
)
PSX_ROMS_FOR_REGION = {
    "JA": ["scph5500.bin", PSX_SCPH5500_BIN.sha1],
    "US": ["scph5501.bin", PSX_SCPH5501_BIN.sha1],
    "EU": ["scph5502.bin", PSX_SCPH5502_BIN.sha1],
}
PSX_MODEL_NTSC = "ntsc"
PSX_MODEL_NTSC_J = "ntsc-j"
PSX_MODEL_PAL = "pal"


class PlayStationPlatform(Platform):
    # FIXME: Move to init instead
    PLATFORM_NAME = PSX_PLATFORM_NAME

    def driver(self, fsgc):
        return PlayStationMednafenDriver(fsgc)

    def loader(self, fsgc):
        return PlayStationLoader(fsgc)


class PlayStationLoader(SimpleLoader):
    def load_files(self, values):
        self.config["cue_sheets"] = values["cue_sheets"]
        self.config["file_list"] = values["file_list"]
        self.config["sbi_data"] = values["sbi_data"]

    def load_extra(self, values):
        self.config[Option.PSX_MODEL] = values["psx_model"]
        self.config[Option.PSX_PORT_1_TYPE] = values["psx_port_1_type"]
        self.config[Option.PSX_PORT_2_TYPE] = values["psx_port_2_type"]


PSX_PORTS = [
    {
        "description": "Port 1",
        "types": [PSX_GAMEPAD],
        "type_option": "psx_port_1_type",
        "device_option": "psx_port_1",
    },
    {
        "description": "Port 2",
        "types": [PSX_GAMEPAD],
        "type_option": "psx_port_2_type",
        "device_option": "psx_port_2",
    },
]


class PlayStationMednafenDriver(MednafenDriver):
    PORTS = PSX_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = PlayStationHelper(self.options)

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "CIRCLE": "psx.input.port1.gamepad.circle",
                "CROSS": "psx.input.port1.gamepad.cross",
                "TRIANGLE": "psx.input.port1.gamepad.triangle",
                "SQUARE": "psx.input.port1.gamepad.square",
                "L1": "psx.input.port1.gamepad.l1",
                "L2": "psx.input.port1.gamepad.l2",
                "R1": "psx.input.port1.gamepad.r1",
                "R2": "psx.input.port1.gamepad.r2",
                "UP": "psx.input.port1.gamepad.up",
                "DOWN": "psx.input.port1.gamepad.down",
                "LEFT": "psx.input.port1.gamepad.left",
                "RIGHT": "psx.input.port1.gamepad.right",
                "SELECT": "psx.input.port1.gamepad.select",
                "START": "psx.input.port1.gamepad.start",
            }
        elif port == 1:
            return {
                "CIRCLE": "psx.input.port2.gamepad.circle",
                "CROSS": "psx.input.port2.gamepad.cross",
                "TRIANGLE": "psx.input.port2.gamepad.triangle",
                "SQUARE": "psx.input.port2.gamepad.square",
                "L1": "psx.input.port2.gamepad.l1",
                "L2": "psx.input.port2.gamepad.l2",
                "R1": "psx.input.port2.gamepad.r1",
                "R2": "psx.input.port2.gamepad.r2",
                "UP": "psx.input.port2.gamepad.up",
                "DOWN": "psx.input.port2.gamepad.down",
                "LEFT": "psx.input.port2.gamepad.left",
                "RIGHT": "psx.input.port2.gamepad.right",
                "SELECT": "psx.input.port2.gamepad.select",
                "START": "psx.input.port2.gamepad.start",
            }

    def mednafen_system_prefix(self):
        return "psx"

    def game_video_size(self):
        # FIXME
        if self.is_pal():
            size = (320, 240)
        else:
            size = (320, 224)
        return size

    def prepare(self):
        super().prepare()

        # self.init_mednafen_crop_from_viewport()
        self.set_mednafen_aspect(4, 3)
        # We do aspect calculation separately. Must not be done twice.
        # self.emulator.args.extend(["-psx.correct_aspect", "0"])

        self.emulator.args.extend(["-psx.region_autodetect", "0"])
        if self.helper.model() == PSX_MODEL_PAL:
            self.prepare_mednafen_bios(PSX_SCPH5502_BIN, "scph5502.bin")
            self.emulator.args.extend(["-psx.region_default", "eu"])
        elif self.helper.model() == PSX_MODEL_NTSC_J:
            self.prepare_mednafen_bios(PSX_SCPH5500_BIN, "scph5500.bin")
            self.emulator.args.extend(["-psx.region_default", "jp"])
        else:
            self.prepare_mednafen_bios(PSX_SCPH5501_BIN, "scph5501.bin")
            self.emulator.args.extend(["-psx.region_default", "na"])

        if self.options[Option.PSX_PRELOAD] == "1":
            self.emulator.args.extend(["-cd.image_memcache", "1"])

        self.prepare_mednafen_cd_images()

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class PlayStationHelper:
    def __init__(self, options):
        self.options = options

    def model(self):
        model = self.options[Option.PSX_MODEL]
        if model in ["", PSX_MODEL_NTSC]:
            return PSX_MODEL_NTSC
        elif model == PSX_MODEL_PAL:
            return PSX_MODEL_PAL
        elif model == PSX_MODEL_NTSC_J:
            return PSX_MODEL_NTSC_J
        else:
            print("[PSX] Warning: Invalid model:", model)
            return PSX_MODEL_NTSC

    def pal(self):
        return self.model() == PSX_MODEL_PAL



import logging
import os
import sys
import threading
import time

import fsboot

# using this lock to serialize logging from different threads
lock = threading.RLock()


class MultiplexedOutput:
    def __init__(self, *files):
        self.files = files

    def flush(self):
        with lock:
            for f in self.files:
                try:
                    f.flush()
                except Exception:
                    pass

    def isatty(self):
        return False

    def write(self, msg):
        with lock:
            for f in self.files:
                try:
                    f.write(msg)
                except Exception:
                    pass


class FileOutput(object):
    def __init__(self, file_obj):
        self.file = file_obj
        self.new_line = False

    def flush(self):
        return self.file.flush()

    def isatty(self):
        return False

    def write(self, msg):
        if isinstance(msg, str):
            self.write(msg.encode("UTF-8"))
            return
        # FIXME: legacy hack, should be removed in the future
        if b"database_password" in msg:
            return

        if self.new_line:
            elapsed = time.perf_counter() - fsboot.perf_counter_epoch
            self.file.write("{:0.3f} ".format(elapsed).encode("ASCII"))

        self.file.write(msg)
        self.new_line = msg.endswith(b"\n")


class NullOutput(object):
    def flush(self):
        pass

    def isatty(self):
        return False

    def write(self, msg):
        pass


def setup_logging(log_name):
    if sys.platform == "win32" and "FS_FORCE_STDOUT" not in os.environ:
        # noinspection PyUnresolvedReferences
        # import win32console
        if hasattr(sys, "frozen"):
            # or win32console.GetConsoleWindow() == 0:
            sys.stdout = NullOutput()
            sys.stderr = NullOutput()

    # FIXME: remove dependency on fsgs here!
    from fsgs.FSGSDirectories import FSGSDirectories

    logs_dir = FSGSDirectories.get_logs_dir()
    log_file = os.path.join(logs_dir, log_name)
    print("[LOGGING] Logging to", log_file)
    try:
        f = open(log_file, "wb")
    except Exception:
        print("[LOGGING] Could not open log file")
        # use MultiplexedOutput here too, for the mutex handling
        sys.stdout = MultiplexedOutput(sys.stdout)
        sys.stderr = MultiplexedOutput(sys.stderr)
    else:
        sys.stdout = MultiplexedOutput(FileOutput(f), sys.stdout)
        sys.stderr = MultiplexedOutput(FileOutput(f), sys.stderr)

    logging.basicConfig(stream=sys.stdout, level=logging.NOTSET)



from fsbc.settings import Settings
from fsbc.user import get_user_name
from launcher.option import Option

from .launcher_signal import LauncherSignal


class LauncherSettings(object):
    base_dir = ""

    # These are not used any more, and should be removed from settings if
    # found.
    old_keys = {"config_base", "config_refresh"}

    # The main use of default_settings is actually to hide the corresponding
    # options from advanced settings (because there is UI for them, or
    # because the settings are internal)
    default_settings = {
        "__netplay_ready": "",
        "__config_refresh": "",
        "automatic_input_grab": "",
        Option.ARCADE_DATABASE: "",
        "arcade_fullscreen": "",
        Option.ATARI_DATABASE: "",
        "audio_frequency": "",
        "audio_buffer_target_size": "",
        "builtin_configs": "",
        Option.C64_DATABASE: "",
        "config_changed": "0",
        "config_name": "Unnamed Configuration",
        "config_feature": "",
        "config_path": "",
        Option.CPC_DATABASE: "",
        "config_search": "",
        "config_xml_path": "",
        "configurations_dir_mtime": "",
        "database_auth": "",
        "database_email": "",
        "database_feature": "",
        "database_password": "",
        "database_server": "",
        Option.DATABASE_SHOW_ADULT: "",
        Option.DATABASE_SHOW_GAMES: "",
        Option.DATABASE_SHOW_UNPUBLISHED: "",
        "database_username": "",
        "device_id": "",
        Option.DOS_DATABASE: "",
        "fsaa": "",
        "floppy_drive_volume": "",
        Option.FLOPPY_DRIVE_VOLUME_EMPTY: "",
        "fullscreen": "",
        "game_uuid": "",
        "game_list_uuid": "",
        Option.GB_DATABASE: "",
        Option.GBA_DATABASE: "",
        Option.GBC_DATABASE: "",
        "initial_input_grab": "",
        "irc_nick": "",
        "irc_server": "",
        "keep_aspect": "",
        "kickstarts_dir_mtime": "",
        "kickstart_setup": "",
        "language": "",
        "last_cartridge_dir": "",
        "last_cd_dir": "",
        "last_floppy_dir": "",
        "last_hd_dir": "",
        "last_rom_dir": "",
        "last_scan": "",
        "last_tape_dir": "",
        "last_settings_page": "",
        Option.LAUNCHER_CLOSE_BUTTONS: "",
        Option.LAUNCHER_THEME: "",
        "low_latency_vsync": "",
        "maximized": "0",
        "middle_click_ungrab": "",
        "monitor": "",
        "mouse_speed": "",
        Option.NEOGEO_DATABASE: "",
        Option.NES_DATABASE: "",
        "netplay_feature": "1",
        "netplay_tag": "",
        "parent_uuid": "",
        Option.PLATFORMS_FEATURE: "",
        "primary_joystick": "",
        Option.PSX_DATABASE: "",
        # "rtg_scanlines": "",
        "scan_configs": "1",
        "scan_files": "1",
        "scan_roms": "1",
        # "scanlines": "",
        "search_path": "",
        "secondary_joystick": "",
        Option.SMD_DATABASE: "",
        Option.SMD_EMULATOR: "",
        Option.SMS_DATABASE: "",
        Option.SNES_DATABASE: "",
        "stereo_separation": "",
        "swap_ctrl_keys": "",
        "texture_filter": "",
        "texture_format": "",
        Option.TG16_DATABASE: "",
        Option.TGCD_DATABASE: "",
        "__variant_rating": "",
        "video_format": "",
        "video_sync": "",
        "video_sync_method": "",
        Option.WHDLOAD_PRELOAD: "",
        Option.WHDLOAD_SPLASH_DELAY: "",
        Option.WHDLOAD_QUIT_KEY: "",
        # "window_width": "",
        # "window_height": "",
        "zoom": "",
        Option.ZXS_DATABASE: "",
    }

    settings = default_settings.copy()
    # settings_listeners = []

    initialize_from_config = {"fullscreen"}

    @classmethod
    def get(cls, key):
        # return app.settings.get(key, cls.default_settings.get(key, ""))
        return Settings.instance().get(key)

    @classmethod
    def set(cls, key, value):
        Settings.instance()[key] = value

    @classmethod
    def keys(cls):
        return Settings.instance().values.keys()

    @classmethod
    def items(cls):
        return Settings.instance().values.items()

    @classmethod
    def add_listener(cls, listener):
        # cls.settings_listeners.append(listener)
        LauncherSignal.add_listener("setting", listener)

    @classmethod
    def remove_listener(cls, listener):
        # cls.settings_listeners.remove(listener)
        LauncherSignal.remove_listener("setting", listener)

    @classmethod
    def get_irc_nick(cls):
        value = cls.get("irc_nick").strip()
        if not value:
            value = get_user_name()
        # these are probably valid too: \`
        valid_chars = (
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_[]{}|^"
        )
        extra_valid_chars = "0123456789-"
        nick = ""
        for c in value:
            if c in valid_chars:
                nick = nick + c
                if extra_valid_chars:
                    valid_chars += extra_valid_chars
                    extra_valid_chars = ""
        if not nick:
            nick = "User"
        return nick

    @classmethod
    def get_irc_nickserv_pass(cls):
        value = cls.get("irc_nickserv_pass").strip()
        if value:
            return value
        return ""

    @classmethod
    def get_irc_server(cls):
        value = cls.get("irc_server").strip()
        if value:
            return value
        return "irc.fs-uae.net"



import fsui

from ...launcher_config import LauncherConfig


class ConfigCheckBox(fsui.CheckBox):
    def __init__(self, parent, label, config_key):
        fsui.CheckBox.__init__(self, parent, label)
        self.config_key = config_key
        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config(self.config_key, LauncherConfig.get(self.config_key))

    def set_config_handlers(self):
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_changed(self):
        if self.is_checked():
            LauncherConfig.set(self.config_key, "1")
        else:
            LauncherConfig.set(self.config_key, "")

    def on_config(self, key, value):
        if key == self.config_key:
            if value == "1":
                self.check(True)
            else:
                self.check(False)



import fsui
from fsbc.application import app
from launcher.i18n import gettext
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior


class MonitorButton(fsui.ImageButton):
    def __init__(self, parent):
        self.left_icon = fsui.Image("launcher:res/16x16/monitor_left.png")
        self.middle_left_icon = fsui.Image(
            "launcher:res/16x16/monitor_middle_left.png"
        )
        self.middle_right_icon = fsui.Image(
            "launcher:res/16x16/monitor_middle_right.png"
        )
        self.right_icon = fsui.Image("launcher:res/16x16/monitor_right.png")
        super().__init__(parent, self.middle_left_icon)
        self.set_tooltip(
            gettext(
                "Monitor to display FS-UAE on (left, "
                "middle-left, middle-right, right)"
            )
        )
        self.set_min_width(40)
        self.monitor = ""
        SettingsBehavior(self, ["fullscreen", "monitor"])

    def on_fullscreen_setting(self, value):
        # self.set_enabled(value == "1")
        pass

    def on_monitor_setting(self, value):
        if value == "left":
            self.monitor = "left"
            self.set_image(self.left_icon)
        elif value == "middle-right":
            self.monitor = "middle-right"
            self.set_image(self.middle_right_icon)
        elif value == "right":
            self.monitor = "right"
            self.set_image(self.right_icon)
        else:
            self.monitor = "middle-left"
            self.set_image(self.middle_left_icon)

    def on_activate(self):
        if self.monitor == "left":
            app.settings["monitor"] = ""
        elif self.monitor == "middle-left":
            app.settings["monitor"] = "middle-right"
        elif self.monitor == "middle-right":
            app.settings["monitor"] = "right"
        else:
            app.settings["monitor"] = "left"



from __future__ import absolute_import
from __future__ import print_function

import struct
from .block.Block import Block
from .block.UserDirBlock import UserDirBlock
from .block.DirCacheBlock import *
from .ADFSFile import ADFSFile
from .ADFSNode import ADFSNode
from .FileName import FileName
from .FSError import *
from .FSString import FSString
from .MetaInfo import *

class ADFSDir(ADFSNode):
  def __init__(self, volume, parent):
    ADFSNode.__init__(self, volume, parent)
    # state
    self.entries = None
    self.dcache_blks = None
    self.name_hash = None
    self.valid = False
    
  def __repr__(self):
    if self.block != None:
      return "[Dir(%d)'%s':%s]" % (self.block.blk_num, self.block.name, self.entries)
    else:
      return "[Dir]"
  
  def blocks_create_old(self, anon_blk):
    ud = UserDirBlock(self.blkdev, anon_blk.blk_num, self.volume.is_longname)
    ud.set(anon_blk.data)
    if not ud.valid:
      raise FSError(INVALID_USER_DIR_BLOCK, block=anon_blk)
    self.set_block(ud)
    return ud
  
  def _read_add_node(self, blk, recursive):
    hash_chain = None
    node = None
    if blk.valid_chksum and blk.type == Block.T_SHORT:
      # its a userdir
      if blk.sub_type == Block.ST_USERDIR:
        node = ADFSDir(self.volume, self)
        blk = node.blocks_create_old(blk)
        if recursive:
          node.read()
      # its a file
      elif blk.sub_type == Block.ST_FILE:
        node = ADFSFile(self.volume, self)
        blk = node.blocks_create_old(blk)
      # unsupported
      else:
        raise FSError(UNSUPPORTED_DIR_BLOCK, block=blk, extra="Sub_Type: %08x" % blk.sub_type)
    hash_chain = blk.hash_chain
    return hash_chain,node
  
  def _init_name_hash(self):
    self.name_hash = []
    for i in xrange(self.block.hash_size):
      self.name_hash.append([])      
  
  def read(self, recursive=False):
    self._init_name_hash()
    self.entries = [] 
     
    # create initial list with blk_num/hash_index for dir scan
    blocks = []
    for i in xrange(self.block.hash_size):
      blk_num = self.block.hash_table[i]
      if blk_num != 0:
        blocks.append((blk_num,i))
  
    for blk_num,hash_idx in blocks:
      # read anonymous block
      blk = Block(self.blkdev, blk_num)
      blk.read()
      if not blk.valid:
        self.valid = False
        return
      # create file/dir node
      hash_chain,node = self._read_add_node(blk, recursive)
      # store node in entries
      self.entries.append(node)
      # store node in name_hash
      self.name_hash[hash_idx].append(node)
      # follow hash chain
      if hash_chain != 0:
        blocks.append((hash_chain,hash_idx))
    
    # dircaches available?
    if self.volume.is_dircache:
      self.dcache_blks = []
      dcb_num = self.block.extension
      while dcb_num != 0:
        dcb = DirCacheBlock(self.blkdev, dcb_num)
        dcb.read()
        if not dcb.valid:
          self.valid = False
          return
        self.dcache_blks.append(dcb)
        dcb_num = dcb.next_cache
    
  def flush(self):
    if self.entries:
      for e in self.entries:
        e.flush()
    self.entries = None
    self.name_hash = None
  
  def ensure_entries(self):
    if not self.entries:
      self.read()
      
  def get_entries(self):
    self.ensure_entries()
    return self.entries
  
  def has_name(self, fn):
    fn_hash = fn.hash()
    fn_up = fn.get_upper_ami_str()
    node_list = self.name_hash[fn_hash]
    for node in node_list:
      if node.name.get_upper_ami_str() == fn_up:
        return True
    return False
  
  def blocks_create_new(self, free_blks, name, hash_chain_blk, parent_blk, meta_info):
    blk_num = free_blks[0]
    blkdev = self.blkdev
    # create a UserDirBlock
    ud = UserDirBlock(blkdev, blk_num, self.volume.is_longname)
    ud.create(parent_blk, name.get_ami_str(), meta_info.get_protect(), meta_info.get_comment_ami_str(), meta_info.get_mod_ts(), hash_chain_blk)
    ud.write()    
    self.set_block(ud)
    self._init_name_hash()
    return blk_num
  
  def blocks_get_create_num(self):
    # the number of blocks needed for a new (empty) directory
    # -> only one UserDirBlock
    return 1
    
  def _create_node(self, node, name, meta_info, update_ts=True):
    self.ensure_entries()
    
    # make sure a default meta_info is available
    if meta_info == None:
      meta_info = MetaInfo()
      meta_info.set_current_as_mod_time()
      meta_info.set_default_protect()
    # check file name
    fn = FileName(name, is_intl=self.volume.is_intl,is_longname=self.volume.is_longname)
    if not fn.is_valid():
      raise FSError(INVALID_FILE_NAME, file_name=name, node=self)
    # does already exist an entry in this dir with this name?
    if self.has_name(fn):
      raise FSError(NAME_ALREADY_EXISTS, file_name=name, node=self)
    # calc hash index of name
    fn_hash = fn.hash()
    hash_chain = self.name_hash[fn_hash]
    if len(hash_chain) == 0:
      hash_chain_blk = 0
    else:
      hash_chain_blk = hash_chain[0].block.blk_num
      
    # return the number of blocks required to create this node
    num_blks = node.blocks_get_create_num()
    
    # try to find free blocks
    free_blks = self.volume.bitmap.alloc_n(num_blks)
    if free_blks == None:
      raise FSError(NO_FREE_BLOCKS, node=self, file_name=name, extra="want %d" % num_blks)
      
    # now create the blocks for this node
    new_blk = node.blocks_create_new(free_blks, name, hash_chain_blk, self.block.blk_num, meta_info)

    # dircache: create record for this node
    if self.volume.is_dircache:
      ok = self._dircache_add_entry(name, meta_info, new_blk, node.get_size(), update_myself=False)
      if not ok:
        self.delete()
        raise FSError(NO_FREE_BLOCKS, node=self, file_name=name, extra="want dcache")

    # update my dir
    self.block.hash_table[fn_hash] = new_blk 
    self.block.write()
    
    # add node
    self.name_hash[fn_hash].insert(0,node)
    self.entries.append(node)
    
    # update time stamps
    if update_ts:
      self.update_dir_mod_time()
      self.volume.update_disk_time()
    
  def update_dir_mod_time(self):
    mi = MetaInfo()
    mi.set_current_as_mod_time()
    self.change_meta_info(mi)
        
  def create_dir(self, name, meta_info=None, update_ts=True):
    if not isinstance(name, FSString):
      raise ValueError("create_dir's name must be a FSString")
    node = ADFSDir(self.volume, self)
    self._create_node(node, name, meta_info, update_ts)
    return node
  
  def create_file(self, name, data, meta_info=None, update_ts=True):
    if not isinstance(name, FSString):
      raise ValueError("create_file's name must be a FSString")
    node = ADFSFile(self.volume, self) 
    node.set_file_data(data)
    self._create_node(node, name, meta_info, update_ts) 
    return node
  
  def _delete(self, node, wipe, update_ts):
    self.ensure_entries()
    
    # can we delete?
    if not node.can_delete():
      raise FSError(DELETE_NOT_ALLOWED, node=node)
    # make sure its a node of mine
    if node.parent != self:
      raise FSError(INTERNAL_ERROR, node=node, extra="node parent is not me")
    if node not in self.entries:
      raise FSError(INTERNAL_ERROR, node=node, extra="node not in entries")      
    # get hash key
    hash_key = node.name.hash()
    names = self.name_hash[hash_key]
    # find my node
    pos = None
    for i in xrange(len(names)):
      if names[i] == node:
        pos = i
        break
    # hmm not found?!
    if pos == None:
      raise FSError(INTERNAL_ERROR, node=node, extra="node not found in hash chain")
    # find prev and next in hash list
    if pos > 0:
      prev = names[pos-1]
    else:
      prev = None
    if pos == len(names)-1:
      next_blk = 0
    else:
      next_blk = names[pos+1].block.blk_num

    # remove node from the hash chain
    if prev == None:
      self.block.hash_table[hash_key] = next_blk
      self.block.write()
    else:
      prev.block.hash_chain = next_blk
      prev.block.write()

    # remove from my lists
    self.entries.remove(node)
    names.remove(node)

    # remove blocks of node in bitmap
    blk_nums = node.get_block_nums()
    self.volume.bitmap.dealloc_n(blk_nums)
    
    # dircache?
    if self.volume.is_dircache:
      free_blk_num = self._dircache_remove_entry(node.name.name)
    else:
      free_blk_num = None
    
    # (optional) wipe blocks
    if wipe:
      clr_blk = '\0' * self.blkdev.block_bytes
      for blk_num in blk_nums:
        self.blkdev.write_block(blk_num, clr_blk)
      # wipe a potentially free'ed dircache block, too
      if free_blk_num != None:
        self.blkdev.write_block(free_blk_num, clr_blk)
    
    # update time stamps
    if update_ts:
      self.update_dir_mod_time()
      self.volume.update_disk_time()
    
  def can_delete(self):
    self.ensure_entries()
    return len(self.entries) == 0
  
  def delete_children(self, wipe, all, update_ts):
    self.ensure_entries()
    entries = self.entries[:]
    for e in entries:
      e.delete(wipe, all, update_ts)

  def get_entries_sorted_by_name(self):
    self.ensure_entries()
    return sorted(self.entries, key=lambda x : x.name.get_upper_ami_str())
    
  def list(self, indent=0, all=False, detail=False, encoding="UTF-8"):
    ADFSNode.list(self, indent, all, detail, encoding)
    if not all and indent > 0:
      return
    self.ensure_entries()
    es = self.get_entries_sorted_by_name()
    for e in es:
      e.list(indent=indent+1, all=all, detail=detail, encoding=encoding)
    
  def get_path(self, pc, allow_file=True, allow_dir=True):
    if len(pc) == 0:
      return self
    self.ensure_entries()
    for e in self.entries:
      if not isinstance(pc[0], FileName):
        raise ValueError("get_path's pc must be a FileName array")
      if e.name.get_upper_ami_str() == pc[0].get_upper_ami_str():
        if len(pc) > 1:
          if isinstance(e, ADFSDir):
            return e.get_path(pc[1:], allow_file, allow_dir)
          else:
            return None
        else:
          if isinstance(e, ADFSDir):
            if allow_dir:
              return e
            else:
              return None
          elif isinstance(e, ADFSFile):
            if allow_file:
              return e
            else:
              return None
          else:
            return None
    return None
    
  def draw_on_bitmap(self, bm, show_all=False, first=True):
    blk_num = self.block.blk_num
    bm[blk_num] = 'D'
    if show_all or first:
      self.ensure_entries()
      for e in self.entries:
        e.draw_on_bitmap(bm, show_all, False)
    if self.dcache_blks != None:
      for dcb in self.dcache_blks:
        bm[dcb.blk_num] = 'C'

  def get_block_nums(self):
    self.ensure_entries()
    result = [self.block.blk_num]
    if self.volume.is_dircache:
      for dcb in self.dcache_blks:
        result.append(dcb.blk_num)
    return result

  def get_blocks(self, with_data=False):
    self.ensure_entries()
    result = [self.block]
    if self.volume.is_dircache:
      result += self.dcache_blks
    return result

  def get_size(self):
    return 0
  
  def get_size_str(self):
    return "DIR"

  def get_detail_str(self):
    self.ensure_entries()
    if self.entries != None:
      s = "entries=%d" % len(self.entries)
    else:
      s = ""
    if self.dcache_blks != None:
      s += " dcache=%d" % len(self.dcache_blks)
    return s

  # ----- dir cache -----

  def _dircache_add_entry(self, name, meta_info, entry_blk, size, update_myself=True):
    # create a new dircache record
    r = DirCacheRecord(entry=entry_blk, size=size, protect=meta_info.get_protect(), \
                       mod_ts=meta_info.get_mod_ts(), sub_type=0, name=name, comment=meta_info.get_comment())
    return self._dircache_add_entry_int(r, update_myself)
    
  def _dircache_add_entry_int(self, r, update_myself=True):
    r_bytes = r.get_size()
    # find a dircache block with enough space
    found_blk = None
    for dcb in self.dcache_blks:
      free_bytes = dcb.get_free_record_size()
      if r_bytes < free_bytes:
        found_blk = dcb
        break
    # need to create a new one?
    if found_blk == None:
      found_blk = self._dircache_add_block(update_myself)
      if found_blk == None:
        return False
    # add record to block and update it
    found_blk.add_record(r)
    found_blk.write()
    return True
  
  def _dircache_add_block(self, update_myself):
    # allocate block
    blk_nums = self.volume.bitmap.alloc_n(1)
    if blk_nums == None:
      return None
    # setup dir cache block
    dcb_num = blk_nums[0]
    dcb = DirCacheBlock(self.blkdev, dcb_num)
    dcb.create(parent=self.block.blk_num)
    # link new cache block
    if len(self.dcache_blks) == 0:
      self.block.extension = dcb_num
      if update_myself:
        self.block.write()
    else:
      last_dcb = self.dcache_blks[-1]
      last_dcb.next_cache = dcb_num
      last_dcb.write()
    self.dcache_blks.append(dcb)
    return dcb
    
  def _dircache_remove_entry(self, name, update_myself=True):
    # first find entry
    pos = None
    dcb = None
    record = None
    n = len(self.dcache_blks)
    for i in xrange(n):
      dcb = self.dcache_blks[i]
      record = dcb.get_record_by_name(name)
      if record != None:
        pos = i
        break
    if record == None:
      raise FSError(INTERNAL_ERROR, node=self)
    # remove entry from this block
    dcb.remove_record(record)
    # remove whole block?
    if dcb.is_empty():
      # next block following me
      if pos == n-1:
        next = 0
      else:
        next = self.dcache_blks[pos+1].blk_num
      # update block links
      if pos == 0:
        # adjust extension link in this dir node
        self.block.extension = next
        if update_myself:
          self.block.write()
      else:
        # adjust dircache block in front of me
        prev_blk = self.dcache_blks[pos-1]
        prev_blk.next_cache = next
        prev_blk.write()
      # free cache block in bitmap
      blk_num = dcb.blk_num
      self.volume.bitmap.dealloc_n([blk_num])
      return blk_num # return number of just deleted block
    else:
      # update cache block with reduced set of records
      dcb.write()
      return None
    
  def get_dircache_record(self, name):
    if self.dcache_blks != None:
      for dcb in self.dcache_blks:
        record = dcb.get_record_by_name(name)
        if record != None:
          return record
    return None
  
  def update_dircache_record(self, record, rebuild):
    if self.dcache_blks == None:
      return
    # update record
    if rebuild:
      self._dircache_remove_entry(record.name, update_myself=False)
      self._dircache_add_entry_int(record, update_myself=True)
    else:
      # simply re-write the dircache block
      for dcb in self.dcache_blks:
        if dcb.has_record(record):
          dcb.write()
          break
  
  def get_block_usage(self, all=False, first=True):
    num_non_data = 1
    num_data = 0
    if self.dcache_blks != None:
      num_non_data += len(self.dcache_blks)
    if all or first:
      self.ensure_entries()
      for e in self.entries:
        bu = e.get_block_usage(all=all, first=False)
        num_data += bu[0]
        num_non_data += bu[1]
    return (num_data, num_non_data)
  
  def get_file_bytes(self, all=False, first=True):
    size = 0
    if all or first:
      self.ensure_entries()
      for e in self.entries:
        size += e.get_file_bytes(all=all, first=False)
    return size

  def is_dir(self):
    return True



import unittest
import fsui as fsui


class WidgetTestCase(unittest.TestCase):
    # noinspection PyPep8Naming
    def __init__(self, methodName="runTest"):
        unittest.TestCase.__init__(self, methodName=methodName)
        self.dialog = None

    def setup_dialog(self):
        self.dialog = fsui.LegacyDialog()

    def run_dialog(self):
        def run_in_dialog():
            self.dialog.end_modal(True)

        fsui.call_after(run_in_dialog)
        self.dialog.show_modal()
        self.dialog.destroy()



from fsgs.drivers.mednafen.lynxdriver import MednafenLynxDriver
from fsgs.platform import PlatformHandler
from fsgs.platforms.loader import SimpleLoader


class LynxPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Lynx"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return GameGearLoader(fsgs)

    def get_runner(self, fsgs):
        return MednafenLynxDriver(fsgs)


class GameGearLoader(SimpleLoader):
    pass



from __future__ import absolute_import
from __future__ import print_function

from .FSError import *

class ProtectFlags:
  FIBF_DELETE = 1
  FIBF_EXECUTE = 2
  FIBF_WRITE = 4
  FIBF_READ = 8
  FIBF_ARCHIVE = 16
  FIBF_PURE = 32
  FIBF_SCRIPT = 64
  
  flag_txt = "HSPArwed"
  flag_num = len(flag_txt)
  flag_none = 0xf # --------
  empty_string = "-" * flag_num
  
  def __init__(self, mask=0):
    self.mask = mask
    
  def __str__(self):
    txt = ""
    pos = self.flag_num - 1
    m = 1 << pos
    for i in xrange(self.flag_num):
      bit = self.mask & m == m
      show = '-'
      flg = self.flag_txt[i]
      flg_low = flg.lower()
      if bit:
        if flg_low != flg:
          show = flg_low
      else:
        if flg_low == flg:
          show = flg_low
      txt += show
      m >>= 1
      pos -= 1
    return txt
    
  def bin_str(self):
    res = ""
    m = 1 << (self.flag_num - 1)
    for i in xrange(self.flag_num):
      if m & self.mask == m:
        res += "1"
      else:
        res += "0" 
      m >>= 1
    return res

  def short_str(self):
    return str(self).replace("-","")
    
  def parse(self, s):
    if len(s) == 0:
      return
    # allow to add with '+' or sub with '-'
    n = self.flag_txt
    mode = '+'
    self.mask = self.flag_none
    for a in s.lower():
      if a in '+-':
        mode = a
      else:
        mask = None   
        is_low = None
        for i in xrange(self.flag_num):
          flg = self.flag_txt[i]
          flg_low = flg.lower()
          if flg_low == a:
            mask = 1<<(self.flag_num - 1 - i)
            is_low = flg_low == flg
            break
        if mask == None:
          raise FSError(INVALID_PROTECT_FORMAT,extra="char: "+a)
        # apply mask
        if mode == '+':
          if is_low:
            self.mask &= ~mask
          else:
            self.mask |= mask
        else:
          if is_low:
            self.mask |= mask
          else:
            self.mask &= ~mask
  
  def is_set(self, mask):
    return self.mask & mask == 0 # LO active
  def set(self, mask):
    self.mask &= ~mask
  def clr(self, mask):
    self.mask |= mask

  def is_d(self):
    return self.is_set(self.FIBF_DELETE)
  def is_e(self):
    return self.is_set(self.FIBF_EXECUTE)
  def is_w(self):
    return self.is_set(self.FIBF_WRITE)
  def is_r(self):
    return self.is_set(self.FIBF_READ)

if __name__ == '__main__':
  inp = ["h","s","p","a","r","w","e","d"]
  for i in inp:
    p = ProtectFlags() 
    p.parse(i)
    s = str(p)
    if not i in s:
      print(s)

    


from __future__ import absolute_import
from __future__ import print_function

from .block.EntryBlock import EntryBlock
from .block.FileHeaderBlock import FileHeaderBlock
from .block.FileListBlock import FileListBlock
from .block.FileDataBlock import FileDataBlock
from .ADFSNode import ADFSNode
from .FSError import *

class ADFSFile(ADFSNode):
  def __init__(self, volume, parent):
    ADFSNode.__init__(self, volume, parent)
    # state
    self.ext_blk_nums = []
    self.ext_blks = []
    self.data_blk_nums = []
    self.data_blks = []
    self.valid = False
    self.data = None
    self.data_size = 0
    self.total_blks = 0
  
  def __repr__(self):
    return "[File(%d)'%s':%d]" % (self.block.blk_num, self.block.name, self.block.byte_size)
  
  def blocks_create_old(self, anon_blk):
    # create file header block
    fhb = FileHeaderBlock(self.blkdev, anon_blk.blk_num, self.volume.is_longname)
    fhb.set(anon_blk.data)
    if not fhb.valid:
      raise FSError(INVALID_FILE_HEADER_BLOCK, block=anon_blk)
    self.set_block(fhb)

    # retrieve data blocks and size from header
    self.data_blk_nums = fhb.data_blocks[:]
    self.data_size = fhb.byte_size

    # scan for extension blocks
    next_ext = self.block.extension
    while next_ext != 0:
      ext_blk = FileListBlock(self.block.blkdev, next_ext)
      ext_blk.read()
      if not ext_blk.valid:
        raise FSError(INVALID_FILE_LIST_BLOCK, block=ext_blk)
      self.ext_blk_nums.append(next_ext)
      self.ext_blks.append(ext_blk)
      self.data_blk_nums += ext_blk.data_blocks
      next_ext = ext_blk.extension
    
    # now check number of ext blocks
    self.num_ext_blks = self.calc_number_of_list_blks()
    my_num_ext_blks = len(self.ext_blks)
    if my_num_ext_blks != self.num_ext_blks:
      raise FSError(FILE_LIST_BLOCK_COUNT_MISMATCH, node=self, extra="got=%d want=%d" % (my_num_ext_blks, self.num_ext_blks))

    # now check number of data blocks
    self.num_data_blks = self.calc_number_of_data_blks()
    my_num_data_blks = len(self.data_blk_nums)
    if my_num_data_blks != self.num_data_blks:
      raise FSError(FILE_DATA_BLOCK_COUNT_MISMATCH, node=self, extra="got=%d want=%d" % (my_num_data_blks, self.num_data_blks))

    # calc number of total blocks occupied by this file
    self.total_blks = 1 + my_num_ext_blks + my_num_data_blks
    if self.block.comment_block_id != 0:
        self.total_blks += 1

    return fhb
  
  def read(self):
    """read data blocks"""
    self.data_blks = []
    want_seq_num = 1
    total_size = 0
    is_ffs = self.volume.is_ffs
    byte_size = self.block.byte_size
    data = ""
    for blk in self.data_blk_nums:
      if is_ffs:
        # ffs has raw data blocks
        dat_blk = self.volume.blkdev.read_block(blk)
        total_size += len(dat_blk)
        # shrink last read if necessary
        if total_size > byte_size:
          shrink = total_size - byte_size
          dat_blk = dat_blk[:-shrink]
          total_size = byte_size
        data += dat_blk
      else:
        # ofs
        dat_blk = FileDataBlock(self.block.blkdev, blk)
        dat_blk.read()
        if not dat_blk.valid:
          raise FSError(INVALID_FILE_DATA_BLOCK, block=dat_blk, node=self)
        # check sequence number
        if dat_blk.seq_num != want_seq_num:
          raise FSError(INVALID_SEQ_NUM, block=dat_blk, node=self, extra="got=%d wanted=%d" % (dat_blk.seq_num, want_seq_num))
        # store data blocks
        self.data_blks.append(dat_blk)
        total_size += dat_blk.data_size
        data += dat_blk.get_block_data()
      want_seq_num += 1
    # store full contents of file
    self.data = data
    # make sure all went well
    got_size = len(data)
    want_size = self.block.byte_size
    if got_size != want_size:
      raise FSError(INTERNAL_ERROR, block=self.block, node=self, extra="file size mismatch: got=%d want=%d" % (got_size, want_size))
  
  def get_file_data(self):
    if self.data != None:
      return self.data
    self.read()
    return self.data
  
  def flush(self):
    self.data = None
    self.data_blks = None
  
  def ensure_data(self):
    if self.data == None:
      self.read()
  
  def set_file_data(self, data):
    self.data = data
    self.data_size = len(data)
    self.num_data_blks = self.calc_number_of_data_blks()
    self.num_ext_blks = self.calc_number_of_list_blks()
  
  def get_data_block_contents_bytes(self):
    """how many bytes of file data can be stored in a block?"""
    bb = self.volume.blkdev.block_bytes
    if self.volume.is_ffs:
      return bb
    else:
      return bb - 24
  
  def calc_number_of_data_blks(self):
    """given the file size: how many data blocks do we need to store the file?"""
    bb = self.get_data_block_contents_bytes()
    ds = self.data_size
    return (ds + bb -1 ) / bb
  
  def calc_number_of_list_blks(self):
    """given the file size: how many list blocks do we need to store the data blk ptrs?"""
    db = self.calc_number_of_data_blks()
    # ptr per block
    ppb = self.volume.blkdev.block_longs - 56
    # fits in header block?
    if db <= ppb:
      return 0
    else:
      db -= ppb
      return (db + ppb - 1) / ppb 
  
  def blocks_get_create_num(self):
    # determine number of blocks to create
    return 1 + self.num_data_blks + self.num_ext_blks
  
  def blocks_create_new(self, free_blks, name, hash_chain_blk, parent_blk, meta_info):
    # assign block numbers
    fhb_num = free_blks[0]
    # ... for ext
    self.ext_blk_nums = []
    for i in xrange(self.num_ext_blks):
      self.ext_blk_nums.append(free_blks[1+i])
    # ... for data
    off = 1 + self.num_ext_blks
    self.data_blk_nums = []
    for i in xrange(self.num_data_blks):
      self.data_blk_nums.append(free_blks[off])
      off += 1
    
    ppb = self.volume.blkdev.block_longs - 56 # data pointer per block
    
    # create file header block
    fhb = FileHeaderBlock(self.blkdev, fhb_num, self.volume.is_longname)
    byte_size = len(self.data)
    if self.num_data_blks > ppb:
      hdr_blks = self.data_blk_nums[0:ppb]
      hdr_ext = self.ext_blk_nums[0]
    else:
      hdr_blks = self.data_blk_nums
      hdr_ext = 0
      
    fhb.create(parent_blk, name.get_ami_str(), hdr_blks, hdr_ext, byte_size, meta_info.get_protect(), meta_info.get_comment_ami_str(), meta_info.get_mod_ts(), hash_chain_blk)
    fhb.write() 
    self.set_block(fhb)
    
    # create file list (=ext) blocks
    ext_off = ppb
    for i in xrange(self.num_ext_blks):
      flb = FileListBlock(self.blkdev, self.ext_blk_nums[i])
      if i == self.num_ext_blks - 1:
        ext_blk = 0
        blks = self.data_blk_nums[ext_off:]
      else:
        ext_blk = self.ext_blk_nums[i+1]
        blks = self.data_blk_nums[ext_off:ext_off+ppb]
      flb.create(fhb_num, blks, ext_blk)
      flb.write()
      self.ext_blks.append(flb)
      ext_off += ppb
    
    # write data blocks
    self.write()

    self.valid = True    
    return fhb_num
  
  def write(self):
    self.data_blks = []
    off = 0
    left = self.data_size
    blk_idx = 0
    bs = self.get_data_block_contents_bytes()
    is_ffs = self.volume.is_ffs
    while off < self.data_size:
      # number of data block
      blk_num = self.data_blk_nums[blk_idx]
      # extract file data
      size = left
      if size > bs:
        size = bs
      d = self.data[off:off+size]
      if is_ffs:
        # pad block
        if size < bs:
          d += '\0' * (bs-size)
        # write raw block data in FFS
        self.blkdev.write_block(blk_num, d)
      else:
        # old FS: create and write data block
        fdb = FileDataBlock(self.blkdev, blk_num)
        if blk_idx == self.num_data_blks - 1:
          next_data = 0
        else:
          next_data = self.data_blk_nums[blk_idx+1]
        fdb.create(self.block.blk_num, blk_idx+1, d, next_data)
        fdb.write()
        self.data_blks.append(fdb)
      blk_idx += 1
      off += bs
      left -= bs
  
  def draw_on_bitmap(self, bm, show_all=False, first=False):
    bm[self.block.blk_num] = 'H'
    for b in self.ext_blk_nums:
      bm[b] = 'E'
    for b in self.data_blk_nums:
      bm[b] = 'd'

  def get_block_nums(self):
    result = [self.block.blk_num]
    result += self.ext_blk_nums
    result += self.data_blk_nums
    return result
  
  def get_blocks(self, with_data=True):
    result = [self.block]
    result += self.ext_blks
    if with_data:
      self.ensure_data()
      result += self.data_blks
    return result
  
  def can_delete(self):
    return True
  
  def get_size(self):
    return self.data_size
  
  def get_size_str(self):
    return "%8d" % self.data_size

  def get_detail_str(self):
    return "data=%d ext=%d" % (len(self.data_blk_nums), len(self.ext_blk_nums)) 

  def get_block_usage(self, all=False, first=True):
    return (len(self.data_blk_nums), len(self.ext_blk_nums) + 1)

  def get_file_bytes(self, all=False, first=True):
    return self.data_size

  def is_file(self):
    return True



"""
FS-UAE Netplay Game Server
Copyright (C) 2012 Frode Solheim

This library is free software; you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at
your option) any later version.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
"""

import random
import socket
import sys
import threading
import time
import traceback
from collections import deque
from hashlib import sha1


def int_to_bytes(number):
    return bytes(
        [
            (number & 0xFF000000) >> 24,
            (number & 0x00FF0000) >> 16,
            (number & 0x0000FF00) >> 8,
            (number & 0x000000FF),
        ]
    )


def bytes_to_int(m):
    return m[0] << 24 | m[1] << 16 | m[2] << 8 | m[3]


def byte_ord(v):
    # print("byte_ord", v)
    try:
        return v[0]
    except TypeError:
        return v


def byte(v):
    return bytes([v])


SERVER_PROTOCOL_VERSION = 1
MAX_PLAYERS = 9
max_drift = 25
num_clients = 2
port = 25100
host = "0.0.0.0"
game = None
game_password = 0
launch_timeout = 0
framerate = 0.02
netplay_ntsc = 0
server_protocol_version = byte(SERVER_PROTOCOL_VERSION)


def create_game_password(value):
    # for python 2 + 3 compatibility
    # if not isinstance(value, unicode):
    value = value.encode("UTF-8")
    # print(repr(value))
    h = sha1()
    # noinspection SpellCheckingInspection
    h.update(b"FSNP")
    val = b""
    for v in value:
        if byte_ord(v) < 128:
            val += byte(v)
    # print("update:", repr(val))
    h.update(val)
    return bytes_to_int(h.digest()[:4])


MESSAGE_READY = 0
MESSAGE_MEM_CHECK = 5
MESSAGE_RND_CHECK = 6
MESSAGE_PING = 7
MESSAGE_PLAYERS = 8
MESSAGE_PLAYER_TAG_0 = 9
MESSAGE_PLAYER_TAG_1 = 10
MESSAGE_PLAYER_TAG_2 = 11
MESSAGE_PLAYER_TAG_3 = 12
MESSAGE_PLAYER_TAG_4 = 13
MESSAGE_PLAYER_TAG_5 = 14
MESSAGE_PLAYER_TAG_6 = 15
MESSAGE_PLAYER_TAG_7 = 16
MESSAGE_PLAYER_TAG_8 = 17

MESSAGE_PLAYER_PING = 18
MESSAGE_PLAYER_LAG = 19
MESSAGE_SET_PLAYER_TAG = 20
MESSAGE_PROTOCOL_VERSION = 21
MESSAGE_EMULATION_VERSION = 22
MESSAGE_ERROR = 23
MESSAGE_TEXT = 24
MESSAGE_SESSION_KEY = 25

# MESSAGE_MEM_CHECK = 5
# MESSAGE_RND_CHECK = 6
# MESSAGE_PING = 7

MESSAGE_MEM_CHECK_MASK = 0x80000000 | (MESSAGE_MEM_CHECK << 24)
MESSAGE_RND_CHECK_MASK = 0x80000000 | (MESSAGE_RND_CHECK << 24)

ERROR_PROTOCOL_MISMATCH = 1
ERROR_WRONG_PASSWORD = 2
ERROR_CANNOT_RESUME = 3
ERROR_GAME_ALREADY_STARTED = 4
ERROR_PLAYER_NUMBER = 5
ERROR_EMULATOR_MISMATCH = 6
ERROR_CLIENT_ERROR = 7
ERROR_MEMORY_DESYNC = 8
ERROR_RANDOM_DESYNC = 9
ERROR_SESSION_KEY = 10
ERROR_GAME_STOPPED = 99


def create_ext_message(ext, data):
    return 0x80000000 | ext << 24 | (data & 0xFFFFFF)


class Client:
    def __init__(self, sock, address):
        # self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket = sock
        self.messages = []
        self.lock = threading.Lock()
        self.address = address
        self.ready = 0
        self.tag = b"PLY"
        self.player = 0
        self.playing = False
        self.frame = 0
        self.frame_times = [0.0 for _ in range(100)]
        self.lag = 0.0
        self.out_seq = 0
        self.mem_check = [(0, 0) for _ in range(100)]
        self.rnd_check = [(0, 0) for _ in range(100)]
        self.ping_sent_at = 0
        self.pings = deque([0 for _ in range(10)])
        self.pings_sum = 0
        self.pings_avg = 0
        # self.protocol_version = 0
        self.emulator_version = b""
        self.session_key = 0
        self.resume_from_packet = 0

        self.temp_a = 0
        self.temp_b = 0
        self.temp_c = 0

        threading.Thread(target=self.__thread_function).start()

    def send_error_message(self, error_num):
        print(self, "error", error_num)
        message = 0x80000000 | MESSAGE_ERROR << 24 | error_num
        self.send_message(message)

    def send_message(self, message):
        with self.lock:
            # if message == 0x87000000:
            #     #print("queueing %08x" % (message,))
            #     #traceback.print_stack()
            #     self.temp_c += 1
            self.__send_data(int_to_bytes(message))

    def __send_data(self, data):
        # if data[0] == '\x87':
        #     #print("queueing ping")
        #     #traceback.print_stack()
        #     self.temp_c += 1
        self.socket.sendall(data)

    def queue_message(self, message):
        with self.lock:
            # if message == 0x87000000:
            #     #print("queueing %08x" % (message,))
            #     #traceback.print_stack()
            #     self.temp_c += 1
            # print("queueing %08x" % (message,))
            # if message & 0x20000000:
            #     traceback.print_stack()
            self.messages.append(int_to_bytes(message))
            if len(self.messages) == 100:
                self.__send_queued_messages()

    def queue_bytes(self, message):
        with self.lock:
            self.messages.append(message)
            if len(self.messages) == 100:
                self.__send_queued_messages()

    def send_queued_messages(self):
        with self.lock:
            self.__send_queued_messages()

    def __send_queued_messages(self):
        data = b"".join(self.messages)
        self.messages[:] = []
        # print("sending", repr(data))
        self.__send_data(data)

    def initialize_client(self):
        print("initialize", self)

        def read_bytes(num):
            read_data = b""
            for i in range(num):
                read_data = read_data + self.socket.recv(1)
            if not len(read_data) == num:
                raise Exception("did not get {0} bytes".format(num))
            return read_data

        # check header message
        data = read_bytes(4)
        if data == b"PING":
            # connection check only
            self.__send_data(b"PONG")
            return False
        # noinspection SpellCheckingInspection
        if data != b"FSNP":
            print(data)
            # noinspection SpellCheckingInspection
            raise Exception("did not get expected FSNP header")
        # check protocol version
        data = self.socket.recv(1)
        # print(repr(data), repr(server_protocol_version))
        if data != server_protocol_version:
            print("protocol mismatch")
            self.send_error_message(ERROR_PROTOCOL_MISMATCH)
            return False
        # check net play password
        password = bytes_to_int(read_bytes(4))
        if password != game_password:
            print("wrong password")
            self.send_error_message(ERROR_WRONG_PASSWORD)
            return False
        # read emulator version
        self.emulator_version = read_bytes(8)
        # read player number and session key, session key is checked to
        # make sure another client cannot hijack this player slot
        self.session_key = bytes_to_int(b"\0" + read_bytes(3))
        self.player = byte_ord(self.socket.recv(1))
        self.tag = read_bytes(3)
        # get package sequence number
        self.resume_from_packet = bytes_to_int(read_bytes(4))

        error = game.add_client(self)
        if error:
            print(repr(error))
            self.send_error_message(error)
            return False

        message = create_ext_message(MESSAGE_SESSION_KEY, self.session_key)
        self.queue_message(message)
        data = (self.player << 8) | num_clients
        message = create_ext_message(MESSAGE_PLAYERS, data)
        self.queue_message(message)

        game.send_player_tags(self)
        self.send_queued_messages()
        print("initialize done for", self)
        return True

    def __thread_function(self):
        try:
            try:
                if not self.initialize_client():
                    # print("initialize failed for", self)
                    return
                self.receive_loop()
            except Exception:
                traceback.print_exc()
                game.stop = True
        finally:
            try:
                self.socket.close()
            except Exception:
                pass

    def receive_loop(self):
        data = b""
        count = 0
        while not game.stop:
            new_data = self.socket.recv(4 - count)
            if not new_data:
                # FIXME
                return
            data = data + new_data
            count = len(data)
            if count == 4:
                self.on_message(data)
                count = 0
                data = b""

    def send_ping(self):
        with self.lock:
            # print("ping?", self.ping_sent_at)
            if self.ping_sent_at == 0:
                self.temp_a += 1
                self.ping_sent_at = time.time()
                message = int_to_bytes(0x80000000 | (7 << 24))
                self.__send_data(message)
                assert self.ping_sent_at > 0

    def on_ping(self):
        # may not need lock here
        with self.lock:
            self.temp_b += 1
            if self.temp_a != self.temp_b:
                print(self.temp_a, self.temp_b, self.temp_c)
            assert self.ping_sent_at > 0
            t = time.time()
            new = (t - self.ping_sent_at) / 1.0
            # print(t, "-", self.ping_sent_at, "=", new)
            old = self.pings.popleft()
            self.pings.append(new)
            # print(self.pings)
            self.pings_sum = self.pings_sum - old + new
            self.pings_avg = self.pings_sum / len(self.pings)
            self.ping_sent_at = 0

    def on_message(self, m):
        message = bytes_to_int(m)
        # with game.lock:
        #     if not self.playing:
        #         print(self, "is no longer playing/valid, ignoring message")

        if message & 0x80000000:
            # ext message
            command = (message & 0x7F000000) >> 24
            data = message & 0x00FFFFFF
            if command == MESSAGE_MEM_CHECK:
                self.mem_check[self.frame % 100] = (data, self.frame)
            elif command == MESSAGE_RND_CHECK:
                self.rnd_check[self.frame % 100] = (data, self.frame)
            elif command == MESSAGE_PING:
                # print("{0:x}".format(message))
                self.on_ping()
            elif command == MESSAGE_TEXT:
                print("received text command")
                remaining = data
                text = b""
                while not game.stop:
                    part = self.socket.recv(remaining)
                    count = len(part)
                    text += part
                    remaining -= count
                    if remaining == 0:
                        if text.lower() == b"!disconnectall":
                            print("Server termination requested")
                            text = b"Requested to terminate the server"
                            game.stop = True
                        if text.decode("utf-8").lower().startswith("!fps"):
                            try:
                                fps = text.decode("utf-8").lower().split()[1]
                                print("FPS requested: " + fps)
                                game.framerate = 1 / int(fps)
                                text = (
                                    b"FrameRate set to "
                                    + fps.encode("utf-8")
                                    + b"fps"
                                )
                            except (IndexError, ValueError, ZeroDivisionError):
                                print("Invalid FPS value")
                                text = b"FrameRate unchanged"
                        if text.lower() == b"!ntsc":
                            game.framerate = 0.016667  # 1/60
                            text = b"moved server to NTSC Mode (60 FPS)"
                            print("NTSC requested")
                        if text.lower() == b"!pal":
                            game.framerate = 0.020  # 1/50
                            text = b"moved server to PAL Mode (50 FPS)"
                            print("PAL requested")
                        game.add_text_message(self, text)
                        break

        elif message & (1 << 30):
            frame = message & 0x3FFFFFFF
            # print("received frame", frame)
            if frame != self.frame + 1:
                print("error, expected frame", self.frame + 1, "got", frame)
            # print("received frame", frame)
            self.frame = frame
            t = time.time()
            self.frame_times[self.frame % 100] = t
            game_t = game.frame_times[self.frame % 100]
            self.lag = t - game_t

        elif message & (1 << 29):
            game.add_input_event(self, message & 0x00FFFFFF)

        else:
            print("warning: unknown command received %x" % (message,))

    def __str__(self):
        return "<Client {0}:{1} {2}>".format(
            self.player, self.tag, self.address
        )


def create_session_key():
    return random.randint(0, 2**24 - 1)


class Game:
    def __init__(self):
        self.started = False
        self.frame = 0
        self.time = 0
        self.clients = []
        self.frame_times = [0.0 for _ in range(100)]
        self.lock = threading.Lock()
        self.stop = False
        self.session_keys = [0 for _ in range(MAX_PLAYERS)]
        self.emulator_version = b""
        self.verified_frame = -1

        if netplay_ntsc == 1:
            self.framerate = 0.016667  # 1/60
        else:
            self.framerate = framerate  # 1/50

    def __start(self):
        if len(self.clients) != num_clients:
            print("error - cannot start until all players have connected")
            return

        print("{0} clients connected, starting game".format(num_clients))
        self.started = True
        threading.Thread(target=self.__thread_function).start()

    def add_client(self, client):
        with self.lock:
            if client.player == 0xFF:
                if client.resume_from_packet != 0:
                    return ERROR_CLIENT_ERROR
                if self.started:
                    return ERROR_GAME_ALREADY_STARTED
                client.player = len(self.clients)
                if client.player == 0:
                    self.emulator_version = client.emulator_version
                else:
                    if self.emulator_version != client.emulator_version:
                        return ERROR_EMULATOR_MISMATCH
                client.session_key = create_session_key()
                self.session_keys[client.player] = client.session_key
                self.clients.append(client)
                client.playing = True
                if not self.started:
                    if len(self.clients) == num_clients:
                        self.__start()
            else:
                if client.player >= len(self.clients):
                    return ERROR_PLAYER_NUMBER
                if self.session_keys[client.player] != client.session_key:
                    return ERROR_SESSION_KEY

                # FIXME: must transfer settings for resuming to work
                # old_client = self.clients[client.player]

                self.clients[client.player] = client
                client.playing = True

                if client.resume_from_packet > 0:
                    # cannot resume yet...
                    print("cannot resume at packet", client.resume_from_packet)
                    return ERROR_CANNOT_RESUME
        return 0
        # FIXME: start the game..?

    def __thread_function(self):
        try:
            self.__game_loop()
        except Exception:
            traceback.print_exc()
            self.stop = True

    # noinspection PyMethodMayBeStatic
    def __send_player_tags(self, send_to_client):
        for i, client in enumerate(game.clients):
            data = bytes_to_int(b"\0" + client.tag)
            message = create_ext_message(MESSAGE_PLAYER_TAG_0 + i, data)
            send_to_client.queue_message(message)

    def send_player_tags(self, client):
        with game.lock:
            self.__send_player_tags(client)

    def __game_loop(self):
        with self.lock:
            for client in self.clients:
                self.__send_player_tags(client)
        print("game loop running")
        self.time = time.time()
        while True:
            if self.stop:
                print("stopping game loop")
                # try to send error message to all players
                with self.lock:
                    for client in self.clients:
                        try:
                            client.send_error_message(ERROR_GAME_STOPPED)
                        except Exception:
                            traceback.print_exc()
                return
            self.__game_loop_iteration()

    def __game_loop_iteration(self):
        # FIXME: higher precision sleep?
        target_time = self.time + self.framerate
        t2 = time.time()
        diff = target_time - t2
        sleep = diff - 0.001
        if sleep > 0.0:
            # print(sleep)
            time.sleep(sleep)
        self.time = target_time
        while time.time() < target_time:
            # busy-loop until target time
            pass

        with self.lock:
            if self.frame % 100 == 0:
                self.__send_status()
            self.frame += 1
            self.frame_times[self.frame % 100] = time.time()
            message = (1 << 30) | self.frame
            self.__send_to_clients(message, True)
            if self.frame % 10 == 0:
                for client in self.clients:
                    client.send_ping()
            if self.frame % 200 == 0:
                self.__print_status()

        self.__check_game()

    def __check_game(self):
        oldest_frame, frames = self.__check_frame_status()
        while oldest_frame > self.verified_frame:
            self.check_synchronization(self.verified_frame + 1)
            self.verified_frame += 1
        diff = self.frame - oldest_frame
        if diff <= max_drift:
            # clients have not drifted too far
            return
        first = True
        count = 0
        while diff > 0 and not self.stop:
            if first:
                first = False
                print("---", self.frame, "acked", frames)
            elif count % 100 == 0:
                print("   ", self.frame, "acked", frames)
            time.sleep(0.02)
            oldest_frame, frames = self.__check_frame_status()
            diff = self.frame - oldest_frame
            count += 1
        self.time = time.time() - 0.02

    def __check_frame_status(self):
        oldest_frame = self.frame
        frames = []
        with self.lock:
            for client in self.clients:
                af = client.frame
                if af < oldest_frame:
                    oldest_frame = af
                frames.append(af)
        return oldest_frame, frames

    def __print_status(self):
        for i, client in enumerate(self.clients):
            print(
                "{0} f {1:6d} p avg: {2:3d} {3:3d}".format(
                    i,
                    client.frame,
                    int(client.pings_avg * 1000),
                    int(client.lag * 1000),
                )
            )

    def __send_status(self):
        for i, client in enumerate(self.clients):
            v = int(client.lag * 1000) & 0x0000FFFF
            message = 0x80000000 | MESSAGE_PLAYER_LAG << 24 | i << 16 | v
            self.__send_to_clients(message)
            v = int(client.pings_avg * 1000) & 0x0000FFFF
            message = 0x80000000 | MESSAGE_PLAYER_PING << 24 | i << 16 | v
            self.__send_to_clients(message)

    def add_input_event(self, client, input_event):
        if not self.started:
            # game was not started - ignoring input event
            print(
                "game not started, ignoring input event {0:08x}".format(
                    input_event
                )
            )
            return
        with self.lock:
            if not client.playing:
                print("client", client, "is no longer valid, dropping msg")
                return
            # for now, just broadcast out again to all clients
            message = (1 << 29) | input_event
            self.__send_to_clients(message)

    def add_text_message(self, from_client, text):
        print("add text message")
        with self.lock:
            for client in self.clients:
                # if from_client == client:
                #     continue
                message = (
                    0x80000000
                    | MESSAGE_TEXT << 24
                    | from_client.player << 16
                    | len(text)
                )
                message = int_to_bytes(message) + text
                print("send", repr(message), "to", client)
                client.queue_bytes(message)

    def send_to_clients(self, message, force_send=False):
        # using a lock here to ensure that send_to_clients can
        # be called from multiple threads, but all clients will
        # still get messages in expected order
        with self.lock:
            self.__send_to_clients(message, force_send)

    def __send_to_clients(self, message, force_send=False):
        for client in self.clients:
            # print("send", message, "to", client)
            client.queue_message(message)
            if force_send:
                client.send_queued_messages()

    # noinspection PyMethodMayBeStatic
    def check_synchronization(self, check_frame):
        # FIXME: MOVE TO GAME CLASS
        # FIXME: ONLY RUN ONCE PER FRAME, not once per frame * clients
        # if self.frame == 0:
        #     return
        with game.lock:
            # print("received mem_check", data)
            # check_frame = self.frame - 1
            if check_frame < 0:
                return
            index = check_frame % 100
            mem_check_data = []
            rnd_check_data = []
            for client in game.clients:
                if client.frame <= check_frame:
                    # cannot check yet
                    return
                # print(client, client.frame, client.mem_check)
                mem_check_data.append((client, client.mem_check[index]))
                rnd_check_data.append((client, client.rnd_check[index]))
            check = mem_check_data[0][1]
            for client, data in mem_check_data:
                if check != data:
                    print("mem_check failed for frame", check_frame)
                    for c, d in mem_check_data:
                        print("* {0:08x}  f {1:05d} {2}".format(d[0], d[1], c))
                    for c in game.clients:
                        c.send_error_message(ERROR_MEMORY_DESYNC)
                        c.send_queued_messages()
                    raise Exception("mem check failed")
            check = rnd_check_data[0][1]
            for client, data in rnd_check_data:
                if check != data:
                    print("rnd_check failed for frame", check_frame)
                    for c, d in rnd_check_data:
                        print("* {0:08x}  f {1:05d} {2}".format(d[0], d[1], c))
                    for c in game.clients:
                        c.send_error_message(ERROR_RANDOM_DESYNC)
                        c.send_queued_messages()
                    raise Exception("rnd check failed")


address_map = {}


def accept_client(server_socket):
    global last_keep_alive_time

    server_socket.settimeout(1)
    client_socket, client_address = server_socket.accept()
    # client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDLOWAT, 4)
    client_socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
    client_socket.settimeout(None)
    client = Client(client_socket, client_address)
    # client.player = len(game.clients) + 1
    print("Client connected", client)
    last_keep_alive_time = time.time()

    # game.add_client(client)
    # client.start()
    # if game.can_start():
    #     game.start()


# stop_accepting = False


def accept_thread(server_socket):
    while not game.stop:
        try:
            accept_client(server_socket)
        except socket.timeout:
            pass
        except Exception:
            traceback.print_exc()
            time.sleep(1.0)


def _run_server():
    global game
    global last_keep_alive_time
    game = Game()
    address = (host, port)
    server_socket = socket.socket()
    server_socket.bind(address)
    server_socket.listen(10)
    print("listening")
    sys.stdout.flush()
    if address[0] != "0.0.0.0":
        print("server listening on", address[0], "port", address[1])
    else:
        print("server listening on port", address[1])
    print("want", num_clients, "client(s)")
    if num_clients > MAX_PLAYERS:
        print("ERROR: max clients are", MAX_PLAYERS)
    threading.Thread(target=accept_thread, args=(server_socket,)).start()
    last_keep_alive_time = time.time()
    while not game.started:
        time.sleep(0.1)
        if launch_timeout:
            t2 = time.time()
            if t2 - last_keep_alive_time > launch_timeout:
                print("game not started yet, aborting (timeout)")
                game.stop = True
                return
                # sys.exit()
    print("game started")
    while not game.stop:
        time.sleep(0.1)


def run_server():
    try:
        _run_server()
    except Exception:
        traceback.print_exc()
    except KeyboardInterrupt:
        traceback.print_exc()
    game.stop = True


def main():
    global \
        port, \
        num_clients, \
        game_password, \
        launch_timeout, \
        netplay_ntsc, \
        framerate
    for arg in sys.argv:
        if arg.startswith("--"):
            parts = arg[2:].split("=", 1)
            if len(parts) == 2:
                key, value = parts
                key = key.lower()
                if key == "port":
                    port = int(value)
                elif key == "players":
                    num_clients = int(value)
                elif key == "password":
                    # game_password = crc32(value) & 0xffffffff
                    game_password = create_game_password(value)
                    # print("game password (numeric) is", game_password)
                elif key == "netplay-ntsc":
                    print("Server launching in NTSC mode")
                    netplay_ntsc = int(value)
                elif key == "framerate":
                    framerate = int(value)
                    if framerate == 0:
                        framerate = 0.02
                    else:
                        framerate = 1 / int(value)
                elif key == "launch-timeout":
                    launch_timeout = int(value)
    run_server()


if __name__ == "__main__":
    main()



import fsui
from fsbc import settings
from fsgs import openretro
from fsgs.platform import Platform
from launcher.i18n import gettext
from launcher.option import Option
from launcher.settings.option_ui import OptionUI
from launcher.settings.platformsettingsdialog import PlatformSettingsDialog
from launcher.settings.settings_page import SettingsPage
from launcher.ui.IconButton import IconButton


class GamePlatformsSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("database-settings", "pkg:workspace")
        title = gettext("Game Platforms")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.hori_layout = None
        self.hori_counter = 0

        if openretro or settings.get(Option.PLATFORMS_FEATURE) == "1":
            # self.add_section(gettext("Game Databases"))

            label = fsui.MultiLineLabel(
                self,
                gettext(
                    "Note: This is an experimental feature. "
                    "Additional plugins are needed."
                ),
                640,
            )
            self.layout.add(label, margin_top=20, margin_bottom=20)

            self.add_database_option(
                Platform.CPC, Option.CPC_DATABASE, "Amstrad CPC"
            )
            self.add_database_option(
                Platform.ARCADE, Option.ARCADE_DATABASE, "Arcade"
            )
            self.add_database_option(
                Platform.A7800, Option.A7800_DATABASE, "Atari 7800"
            )
            self.add_database_option(
                Platform.ATARI, Option.ATARI_DATABASE, "Atari ST"
            )
            self.add_database_option(
                Platform.C64, Option.C64_DATABASE, "Commodore 64"
            )
            self.add_database_option(Platform.DOS, Option.DOS_DATABASE, "DOS")
            self.add_database_option(
                Platform.GB, Option.GB_DATABASE, "Game Boy"
            )
            self.add_database_option(
                Platform.GBA, Option.GBA_DATABASE, "Game Boy Advance"
            )
            self.add_database_option(
                Platform.GBC, Option.GBC_DATABASE, "Game Boy Color"
            )
            self.add_database_option(
                Platform.SMS, Option.SMS_DATABASE, "Master System"
            )
            self.add_database_option(
                Platform.SMD, Option.SMD_DATABASE, "Mega Drive"
            )
            self.add_database_option(
                Platform.NEOGEO, Option.NEOGEO_DATABASE, "Neo-Geo"
            )
            self.add_database_option(
                Platform.NES, Option.NES_DATABASE, "Nintendo"
            )
            self.add_database_option(
                Platform.PSX, Option.PSX_DATABASE, "PlayStation"
            )
            self.add_database_option(
                Platform.SNES, Option.SNES_DATABASE, "Super Nintendo"
            )
            self.add_database_option(
                Platform.TG16, Option.TG16_DATABASE, "TurboGrafx-16"
            )
            self.add_database_option(
                Platform.TGCD, Option.TGCD_DATABASE, "TurboGrafx-CD"
            )
            self.add_database_option(
                Platform.ZXS, Option.ZXS_DATABASE, "ZX Spectrum"
            )

            # label = fsui.MultiLineLabel(
            #     self, gettext(
            #         "Note: Support for additional game databases is an "
            #         "experimental feature and does not provide the "
            #         "same level of maturity as Amiga/CDTV/CD32. "
            #         "Also, additional plugins are needed to play the "
            #         "games."), 640)
            # self.layout.add(label, margin_top=20)

    def add_database_option(self, platform, name, description=""):
        self.options_on_page.add(name)
        group = OptionUI.create_group(
            self, name, description=description, help_button=False
        )

        if self.hori_counter % 2 == 0:
            self.hori_layout = fsui.HorizontalLayout()
            self.layout.add(
                self.hori_layout,
                fill=True,
                margin_top=10,
                margin_bottom=10,
                margin_left=-10,
                margin_right=-10,
            )

        self.hori_layout.add(
            group,
            fill=True,
            expand=-1,
            margin=10,
            margin_top=0,
            margin_bottom=0,
        )
        self.hori_layout.add(
            PlatformSettingsButton(self, platform), margin_right=10
        )

        if self.hori_counter % 2 == 0:
            self.hori_layout.add_spacer(0)
        self.hori_counter += 1


class PlatformSettingsButton(IconButton):
    def __init__(self, parent, platform):
        super().__init__(parent, "16x16/settings.png")
        self.platform = platform
        self.enable(
            len(PlatformSettingsDialog.option_list_for_platform(platform)) > 0
        )

    def on_activate(self):
        PlatformSettingsDialog.open(self.window, self.platform)



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.ui.config.configdialog import ConfigDialog
from launcher.ui.config.configpanel import ConfigPanel
from launcher.ui.IconButton import IconButton


class AdditionalConfigPanel(ConfigPanel):
    def __init__(self, parent):
        super().__init__(parent)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        heading_label = fsui.HeadingLabel(
            self, gettext("Additional Configuration")
        )
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(0, expand=True)
        hori_layout.add(CustomConfigButton(self), margin_right=10)
        self.layout.add_spacer(0)

        amiga_panel = fsui.Panel(self)
        amiga_panel.layout = fsui.VerticalLayout()
        self.layout.add(amiga_panel, fill=True)
        self.add_amiga_option(Option.CPU, parent=amiga_panel)
        self.add_amiga_option(Option.JIT_COMPILER, parent=amiga_panel)
        self.add_amiga_option(Option.FLOPPY_DRIVE_SPEED, parent=amiga_panel)
        self.add_amiga_option(
            Option.FLOPPY_DRIVE_VOLUME_EMPTY, parent=amiga_panel
        )
        self.add_amiga_option(Option.FREEZER_CARTRIDGE, parent=amiga_panel)
        self.add_amiga_option(Option.DONGLE_TYPE, parent=amiga_panel)

    #     AmigaShowBehavior(amiga_panel)
    #     ConfigBehavior(self, [Option.PLATFORM])
    #
    # def on_platform_config(self, _):
    #     self.layout.update()


class CustomConfigButton(IconButton):
    def __init__(self, parent):
        super().__init__(parent, "16x16/settings.png")
        self.activated.connect(self.__activated)

    def __activated(self):
        ConfigDialog.run(self.window)



import traceback
import weakref

from fsbc.system import macosx

# from fsui import default_window_center, default_window_parent
from fsui.qt import QMainWindow, QObject, QSignal, QWidget, Qt, qvalue, qflags
from fsui.qt import QPainter, QPoint, QRect, QGuiApplication
from fsui.qt.Image import Image
from fsui.qt.qt import init_qt, QEvent

# noinspection PyProtectedMember
from fsui.qt.helpers import QParent
import fsboot

windows = set()
_use_fws = False


class Margins:
    def __init__(self):
        self.top = 0
        self.left = 0
        self.right = 0
        self.bottom = 0

    def set(self, value):
        self.top = value
        self.left = value
        self.right = value
        self.bottom = value


# noinspection PyProtectedMember
class RealWindow(QMainWindow):
    def __init__(
        self,
        parent,
        child,
        minimizable=True,
        maximizable=True,
        border=True,
        below=False,
    ):
        super().__init__(parent)
        self.margins = Margins()

        flags = Qt.WindowType.Window
        if macosx:
            flags &= ~Qt.WindowType.WindowFullscreenButtonHint

        if border:
            flags |= Qt.WindowType.CustomizeWindowHint
            flags |= Qt.WindowType.WindowCloseButtonHint
            flags |= Qt.WindowType.WindowTitleHint
            if minimizable:
                flags |= Qt.WindowType.WindowMinimizeButtonHint
            if maximizable:
                flags |= Qt.WindowType.WindowMaximizeButtonHint
            # else:
            #     flags &= ~Qt.WindowType.WindowMaximizeButtonHint
        else:
            flags |= Qt.WindowType.FramelessWindowHint
            flags |= Qt.WindowType.NoDropShadowWindowHint
            if below:
                flags |= Qt.WindowType.WindowStaysOnBottomHint
        self.setWindowFlags(flags)
        # self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)

        # self._child = weakref.ref(child)
        self._child = child
        self.__already_closed = False
        self._size_set = False

        # self.destroyed.connect(self.__on_destroyed)

    @property
    def child(self):
        # return self._child()
        return self._child

    def resize_to(self, size):
        print("RealWindow.resize_to", size[0], size[1])
        x = self.margins.left
        y = self.margins.top
        width = size[0] - self.margins.left - self.margins.right
        height = size[1] - self.margins.top - self.margins.bottom
        assert isinstance(self.child._real_widget, QWidget)
        if self.child.title_panel is not None:
            if self.child.title_panel_visible:
                title_height = self.child.title_panel.height()
                self.child.title_panel.set_position_and_size(
                    (x, y), (width, title_height)
                )
                y += title_height
                height -= title_height
        self.child._real_widget.setGeometry(x, y, width, height)

    def set_size(self, size):
        self._size_set = True
        print("RealWindow.set_size", size)
        width = size[0] + self.margins.left + self.margins.right
        height = size[1] + self.margins.top + self.margins.bottom
        if self.child.title_panel is not None:
            if self.child.title_panel_visible:
                title_height = self.child.title_panel.height()
                height += title_height
        print("RealWindow.set_size final size", (width, height))
        self.resize(width, height)

    def position(self):
        return self.pos().x(), self.pos().y()

    def set_position(self, position):
        self.move(position[0], position[1])

    def width(self):
        return super().width() - self.margins.left - self.margins.right

    def height(self):
        height = super().height() - self.margins.top - self.margins.bottom
        if self.child.title_panel is not None:
            if self.child.title_panel_visible:
                height -= 40
        return height

    def closeEvent(self, event):
        print(str(self) + ".closeEvent")
        event.accept()
        self.__on_closed()

    def showEvent(self, event):
        event.accept()
        self.child.shown.emit()

    def resizeEvent(self, event):
        # event.accept()
        size = event.size()
        print("RealWindow.resizeEvent", size.width(), size.height())
        print(size, "vs", self.size())
        # According to docs, event.size() should be same as
        # QWidget.size(), but this isn't always the case :-/.
        size = self.size()
        self.resize_to((size.width(), size.height()))

    def changeEvent(self, event):
        if event.type() == QEvent.Type.WindowStateChange:
            # print("RealWindow.changeEvent(WindowStateChange)")
            self.child.on_resize()
        QMainWindow.changeEvent(self, event)

    #         size = self.size()
    #         self.resize_to((size.width(), size.height()))

    # def __on_destroyed(self, _):
    #     print("__on_destroyed")
    #     self.child.destroyed.emit()

    def __on_closed(self):
        try:
            print("RealWindow.__on_closed")
            if self.__already_closed:
                print("Looks like a duplicate event, ignoring this one")
                return
            print("closed.emit")
            self.child.closed.emit()
            self.child.on_close()
            self.__already_closed = True
        except Exception:
            traceback.print_exc()

    def is_maximized(self):
        return (
            qflags(self.windowState()) & qvalue(Qt.WindowState.WindowMaximized)
            != 0
        )

    def is_fullscreen(self):
        return (
            qflags(self.windowState()) & qvalue(Qt.WindowState.WindowFullScreen)
            != 0
        )

    def restore_margins(self):
        self.margins.set(0)

    def set_maximized(self, maximize=True, geometry=None):
        print("set_maximized", maximize)
        if maximize:
            self.margins.set(0)
            # if geometry is not None:
            #     print("set_maximized geometry", geometry)
            #     self.setGeometry(*geometry)
            # else:
            #     if not self._size_set:
            #         self._size_set = True
            #         print("resizing to 1, 1")
            #         self.resize(1, 1)
            # self.resize(1, 1)
            # self.resize(1920, 1080)
            self.showMaximized()
            print(
                "size after showMaximized",
                (self.size().width(), self.size().height()),
            )
        else:
            self.restore_margins()
            self.setWindowState(Qt.WindowState.WindowNoState)

    def set_fullscreen(self, fullscreen=True, geometry=None):
        print("set_fullscreen", fullscreen)
        if fullscreen:
            self.margins.set(0)
            # Workaround, without setting size to something initially,
            # showFullScreen sometimes fails to update size on Linux, causing
            # only a small part of the screen to be filled.
            if geometry is not None:
                print("set_fullscreen geometry", geometry)
                self.setGeometry(*geometry)
            else:
                if not self._size_set:
                    self._size_set = True
                    print("resizing to 1, 1")
                    self.resize(1, 1)
            self.showFullScreen()
            print(
                "size after showFullScreen",
                (self.size().width(), self.size().height()),
            )
        else:
            self.restore_margins()
            self.setWindowState(Qt.WindowState.WindowNoState)

    def minimize(self):
        self.setWindowState(Qt.WindowState.WindowMinimized)


# noinspection PyProtectedMember
class RealWidget(QWidget):
    def __init__(self, parent, owner):
        super().__init__(parent)
        self.setAutoFillBackground(True)
        self._owner = weakref.ref(owner)
        self._centered_on_initial_show = False

    def owner(self):
        return self._owner()

    def resizeEvent(self, event):
        self.owner().on_resize()

    def showEvent(self, _):
        if self.owner().center_on_show:
            if not self._centered_on_initial_show:
                if self.parent():
                    self.owner().center_on_parent()
                self._centered_on_initial_show = True

        self.owner().set_initial_size_from_layout()
        self.owner().on_resize()


# noinspection PyPep8Naming
class Window(QObject):
    shown = QSignal()
    closed = QSignal()

    def __init__(
        self,
        parent,
        title="",
        border=True,
        minimizable=True,
        maximizable=True,
        separator=True,
        menu=False,
        native=None,
        header=True,
        below=False,
        closable=True,
        color=None,
        **_,
    ):
        init_qt()
        super().__init__()

        # if parent is None and len(default_window_parent) > 0:
        #     parent = default_window_parent[-1]
        #     print("using default parent", parent)

        # FIXME
        self._window = weakref.ref(self)

        if native is None:
            if fsboot.get("fws") == "1":
                native = False
            else:
                native = not _use_fws
        if native:
            self._real_window = RealWindow(
                QParent(parent, True),
                self,
                minimizable=minimizable,
                maximizable=maximizable,
                below=below,
                border=border,
            )
        else:
            self._real_window = FwsWindow(
                QParent(parent, True), self, below=below
            )

        self._real_widget = RealWidget(self._real_window, self)

        # Widget.__init__(self, parent)
        # self.init_widget(parent)

        # MixinBase.__init__(self)

        self.set_title(title)

        self.layout = None
        self._size_specified = False
        self.close_listeners = []
        # _windows.add(self)
        self.destroyed.connect(self.__destroyed)

        self._real_window.setAttribute(
            Qt.WidgetAttribute.WA_DeleteOnClose, True
        )

        # if not border:
        #     self.setWindowFlags(Qt.WindowType.FramelessWindowHint |
        #                         Qt.WindowType.NoDropShadowWindowHint)
        #     # self.setWindowFlags(Qt.WindowType.FramelessWindowHint)

        self.center_on_show = True
        self._centered_on_initial_show = False
        if hasattr(self, "accepted"):
            self.accepted.connect(self.__accepted)
        if hasattr(self, "rejected"):
            self.rejected.connect(self.__rejected)

        # Add a keep-alive reference
        print("Adding window reference", self)
        windows.add(self)
        self.closed.connect(self.__closed)

        self.title_panel = None
        self.title_panel_visible = False
        if not native:
            self.real_window()._window = weakref.ref(self)
            from workspace.ui.theme import WorkspaceTheme
            from workspace.ui.window import WindowHeader

            if header:
                self.title_panel = WindowHeader(
                    self.real_window(),
                    menu=menu,
                    minimizable=minimizable,
                    maximizable=maximizable,
                    separator=separator,
                    closable=closable,
                    background=color,
                )
            self.title_panel_visible = True
            self.set_background_color(
                WorkspaceTheme.instance().window_background
            )

    def __closed(self):
        print("Removing window reference", self)
        windows.remove(self)

    @property
    def window(self):
        return self

    @property
    def visible(self):
        return self._real_window.isVisible()

    @visible.setter
    def visible(self, visible):
        self._real_window.setVisible(visible)

    def alert(self, msecs=0):
        init_qt().alert(self._real_window, msecs)

    def close(self):
        self._real_window.close()

    def real_window(self):
        return self._real_window

    def real_widget(self):
        return self._real_widget

    # DEPRECATED
    def get_container(self):
        return self.real_widget()

    # FIXME: Is this used?
    def center_on_initial_show(self):
        if self._centered_on_initial_show:
            return
        if self.layout and not self._size_specified:
            self.set_size(self.layout.get_min_size())
        self.on_resize()
        self.center_on_parent()
        self._centered_on_initial_show = True

    def __destroyed(self):
        print(str(self) + ".__destroyed")

    def add_close_listener(self, function):
        # self.close_listeners.append(function)
        self.closed.connect(function)

    def top_level(self):
        return self

    def set_icon(self, icon):
        self.real_window().setWindowIcon(icon.qicon())

    def position(self):
        # position = self._real_window.position()
        # return position.x(), position.y()
        return self._real_window.position()

    def set_position(self, position):
        self._real_window.set_position(position)

    #    def set_position_and_size(self, position, size):
    #        self.widget().setGeometry(position[0], position[1], size[0], size[1])

    def size(self):
        return self.width(), self.height()

    def width(self):
        return self._real_window.width()

    def height(self):
        return self._real_window.height()

    def set_size(self, size):
        self._size_specified = True
        # self.SetClientSize(size)
        # print("FIXME:\n\nDialog.set_size")
        # self.resize(size[0], size[1])
        # import traceback
        # traceback.print_stack()
        if size[0] == 0 or size[1] == 0:
            print("Window.set_size ignoring size", size)
            return
        print("Window.set_size", size)
        self._real_window.set_size(size)
        self._real_widget.resize(size[0], size[1])

    def set_initial_size_from_layout(self):
        if self.layout and not self._size_specified:
            self.set_size_from_layout()

    def set_size_from_layout(self):
        size = self.layout.get_min_size()
        print("set_size_from_layout, size =", size)
        self.set_size(size)

    def title(self):
        return self._real_window.windowTitle()

    def set_title(self, title):
        self._real_window.setWindowTitle(title)

    def show(self, maximized=False, center=False, offset=False):
        if center:
            print("CENTER")
            self.center_on_parent()
            # self._centered_on_initial_show = True
        elif offset:
            self.offset_from_parent((20, 20))
            # self._centered_on_initial_show = True
            self.center_on_show = False
        if maximized:
            self._real_window.showMaximized()
        else:
            self._real_window.show()

    def is_maximized(self):
        # return self._real_window.windowState() == Qt.WindowState.WindowMaximized
        return self._real_window.is_maximized()

    def is_fullscreen(self):
        return self._real_window.is_fullscreen()

    def set_fullscreen(self, fullscreen=True, geometry=None):
        # We must set the size before maximizing, so this isn't done within
        # showMaximized -> ... -> set_initial_size_from_layout -> set_size.
        self.set_initial_size_from_layout()
        if self.title_panel is not None:
            self.title_panel.set_visible(not fullscreen)
            self.title_panel_visible = not fullscreen
        self._real_window.set_fullscreen(fullscreen, geometry)

    def set_maximized(self, maximize=True, geometry=None):
        # We must set the size before maximizing, so this isn't done within
        # showMaximized -> ... -> set_initial_size_from_layout -> set_size.
        self.set_initial_size_from_layout()
        self._real_window.set_maximized(maximize, geometry)

    def minimize(self):
        # self.setWindowState(Qt.WindowState.WindowMinimized)
        return self._real_window.minimize()

    def get_window_center(self):
        position = self.position()
        size = self.size()
        return position[0] + size[0] // 2, position[1] + size[1] // 2

    def center_on_parent(self):
        # import traceback
        # traceback.print_stack()
        self.set_initial_size_from_layout()
        real_parent = self.real_window().parent()
        # print("center_on_parent real_parent = ",
        #       real_parent, default_window_center)
        if real_parent:
            pp = real_parent.x(), real_parent.y()
            ps = real_parent.width(), real_parent.height()
            ss = self.size()
            print(pp, ps, ss)
            self.set_position(
                (pp[0] + (ps[0] - ss[0]) // 2, pp[1] + (ps[1] - ss[1]) // 2)
            )
        # elif len(default_window_center) > 0:
        #     x, y = default_window_center[-1]
        #     ss = self.size()
        #     self.set_position((x - ss[0] // 2, y - ss[1] // 2))

    def offset_from_parent(self, offset):
        self.set_initial_size_from_layout()
        real_parent = self.real_window().parent()
        # print("offset_from_parent real_parent = ",
        #       real_parent, default_window_center)
        if real_parent:
            pp = real_parent.x(), real_parent.y()
            self.set_position((pp[0] + offset[0], pp[1] + offset[1]))

    def center_on_screen(self):
        frame_rect = self._real_window.frameGeometry()
        frame_rect.moveCenter(
            QGuiApplication.primaryScreen().availableGeometry().center()
        )
        self._real_window.move(frame_rect.topLeft())

    def set_background_color(self, color):
        p = self.real_widget().palette()
        p.setColor(self.real_widget().backgroundRole(), color)
        self.real_widget().setPalette(p)

    def is_shown(self):
        return self._real_window.isVisible()

    def raise_and_activate(self):
        self._real_window.raise_()
        self._real_window.activateWindow()

    def on_close(self):
        pass

    def __rejected(self):
        print(str(self) + ".__rejected")
        self.__on_closed()

    def __accepted(self):
        print(str(self) + ".__accepted")
        self.__on_closed()

    def on_resize(self):
        if self.layout:
            self.layout.set_size(self.get_size())
            self.layout.update()

    # DEPRECATED
    def resize(self, width, height):
        self.set_size((width, height))

    # DEPRECATED
    def get_size(self):
        return self.size()

    # DEPRECATED
    def get_window(self):
        return self.top_level()

    # DEPRECATED
    def get_position(self):
        return self.position()

    # DEPRECATED
    def maximized(self):
        return self.is_maximized()

    # DEPRECATED
    def get_parent(self):
        return None

    # DEPRECATED
    def set_icon_from_path(self, _):
        print("FIXME: Window.set_icon_from_path")

    # DEPRECATED
    def get_title(self):
        return self.title()

    # DEPRECATED
    def maximize(self, maximize=True):
        return self.set_maximized(maximize)


class WindowImages:
    def __init__(self):
        self.nw = Image("pkg://workspace.ui/data/window-shadow-nw.png")
        self.n = Image("pkg://workspace.ui/data/window-shadow-n.png")
        self.ne = Image("pkg://workspace.ui/data/window-shadow-ne.png")
        self.e = Image("pkg://workspace.ui/data/window-shadow-e.png")
        self.se = Image("pkg://workspace.ui/data/window-shadow-se.png")
        self.s = Image("pkg://workspace.ui/data/window-shadow-s.png")
        self.sw = Image("pkg://workspace.ui/data/window-shadow-sw.png")
        self.w = Image("pkg://workspace.ui/data/window-shadow-w.png")


class FwsWindow(RealWindow):
    def __init__(self, parent, child, below=False):
        super().__init__(parent, child, border=False, below=below)
        self.restore_margins()
        # self.child = child

        # self.layout = fsui.VerticalLayout(20)
        # self.hori_layout = fsui.HorizontalLayout()
        # self.layout.add(self.hori_layout, fill=True)
        # self.layout.add(child, fill=True, expand=True)

        # self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        from fsui.qt import Qt

        self.setAttribute(Qt.WidgetAttribute.WA_NoSystemBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

    def restore_margins(self):
        self.margins.set(10)

    def paintEvent(self, event):
        # noinspection PyNoneFunctionAssignment
        images = self.images()
        size = self.size()
        w = size.width()
        h = size.height()
        painter = QPainter(self)
        painter.drawImage(QPoint(-10, -10), images.nw.qimage)
        painter.drawImage(QRect(30, -10, w - 60, -10), images.n.qimage)
        painter.drawImage(QPoint(w - 30, -10), images.ne.qimage)
        painter.drawImage(QRect(w - 10, 30, -10, h - 60), images.e.qimage)
        painter.drawImage(QPoint(w - 30, h - 30), images.se.qimage)
        painter.drawImage(QRect(30, h - 10, w - 60, -10), images.s.qimage)
        painter.drawImage(QPoint(-10, h - 30), images.sw.qimage)
        painter.drawImage(QRect(-10, 30, -10, h - 60), images.w.qimage)

    def maximize(self, maximized):
        # FIXME: Better to this via event or something
        if maximized:
            self.layout.set_padding(0)
        else:
            self.layout.set_padding(20)
        super().maximize(maximized)

    _images = None

    @classmethod
    def images(cls):
        if cls._images is None:
            cls._images = WindowImages()
        return cls._images

    @classmethod
    def set_default(cls):
        global _use_fws
        try:
            import workspace
        except ImportError:
            workspace = None
        else:
            _use_fws = True
        del workspace



# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2019  Frode Solheim <frode@solheim.dev>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

PLATFORM_AMIGA = "amiga"
PLATFORM_ARCADE = "arcade"
PLATFORM_ATARI = "atari"
PLATFORM_A2600 = "a2600"
PLATFORM_A5200 = "a5200"
PLATFORM_A7800 = "a7800"
PLATFORM_C64 = "c64"
PLATFORM_CD32 = "cd32"
PLATFORM_CDTV = "cdtv"
PLATFORM_CPC = "cpc"
PLATFORM_DOS = "dos"
PLATFORM_GB = "gb"
PLATFORM_GBA = "gba"
PLATFORM_GBC = "gbc"
PLATFORM_GAME_GEAR_ = "sgg"
# PLATFORM_LYNX = "lynx"
PLATFORM_SMS = "sms"
PLATFORM_SMD = "smd"
PLATFORM_N64 = "n64"
PLATFORM_NES = "nes"
PLATFORM_SNES = "snes"
PLATFORM_TG16 = "tg16"
PLATFORM_ZXS = "zxs"
